# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2007-2015 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from io import StringIO

import songwrite3.plugins as plugins


class ExportPlugin(plugins.ExportPlugin):
  def __init__(self):
    plugins.ExportPlugin.__init__(self, "Songwrite", ["sw.xml"], 1, 1)
    
  def export_to_string(self, song):
    return song.__xml__().getvalue()
    
  def create_ui(self, app): pass
  
ExportPlugin()

class ImportPlugin(plugins.ImportPlugin):
  def __init__(self):
    plugins.ImportPlugin.__init__(self, "Songwrite", ["sw.xml"])
    
  def import_from_string(self, data):
    import songwrite3.stemml as stemml
    return stemml.parse(StringIO(data))
  
  def create_ui(self, app): pass
  
ImportPlugin()



# -*- coding: utf-8 -*-

# SongWrite
# Copyright (C) 2004 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ref : http://anamnese.online.fr/abc/guide_abc.txt


import cStringIO, string, re

import songwrite3.globdef as globdef
import songwrite3.model   as model

_tabnotes = "abcdefghiklmnopqrstuvwxyz"

_notes = {
  "C" : ( "c", "^c",  "d", "^d",  "e",  "f",  "f",  "g", "^g",  "a", "^a",  "b"),
  
  "G" : ( "c", "^c",  "d", "^d",  "e", "_f",  "f",  "g", "^g",  "a", "^a",  "b"),
  "D" : ("_c",  "c",  "d", "^d",  "e", "_f",  "f",  "g", "^g",  "a", "^a",  "b"),
  "A" : ("_c",  "c",  "d", "^d",  "e", "_f",  "f", "_g",  "g",  "a", "^a",  "b"),
  "E" : ("_c",  "c", "_d",  "d",  "e", "_f",  "f", "_g",  "g",  "a", "^a",  "b"),
  "B" : ("_c",  "c", "_d",  "d",  "e", "_f",  "f", "_g",  "g", "_a",  "a",  "b"),
  "F#": ("_c",  "c", "_d",  "d", "_e",  "e",  "f", "_g",  "g", "_a",  "a",  "b"),
  "C#": ( "b",  "c", "_d",  "d", "_e",  "e",  "f", "_g",  "g", "_a",  "a", "_b"),
  
  "F" : ( "c", "^c",  "d", "^d",  "e",  "f", "^f",  "g", "^g",  "a",  "b", "^b"),
  "Bb": ( "c", "^c",  "d",  "e", "^e",  "f", "^f",  "g", "^g",  "a",  "b", "^b"),
  "Eb": ( "c", "^c",  "d",  "e", "^e",  "f", "^f",  "g",  "a", "^a",  "b", "^b"),
  "Ab": ( "c",  "d", "^d",  "e", "^e",  "f", "^f",  "g",  "a", "^a",  "b", "^b"),
  "Db": ( "c",  "d", "^d",  "e", "^e",  "f",  "g", "^g",  "a", "^a",  "b", "^b"),
  "Gb": ("^c",  "d", "^d",  "e", "^e",  "f",  "g", "^g",  "a", "^a",  "b",  "c"),
  "Cb": ("^c",  "d", "^d",  "e",  "f", "^f",  "g", "^g",  "a", "^a",  "b",  "c"),
  }
_duration_2_abc = {
  # Normal duration
  768 : "8",
  384 : "4",
  192 : "2",
   96 : "1",
   48 :  "/2",
   24 :  "/4",
   12 :  "/8",
    6 :  "/16",
    3 :  "/32",
  
  # Doted durations
  576 : "6",
  288 : "3",
  144 : "3/2",
   72 : "3/4",
   36 : "3/8",
   18 : "3/16",
    9 : "3/32",
  
  # Triplets
   32 : "1/3",
   16 : "1/6",
    8 : "1/12",
    4 : "1/24",
  }
_abc_2_duration = dict([(abc, duration) for (duration, abc) in _duration_2_abc.items()])
_abc_2_duration["/"] = _abc_2_duration["/2"]
def duration_2_abc(duration, note = None):
  if note and note.is_start_triplet(): return "(3" + _duration_2_abc[duration]
  return _duration_2_abc[duration]
def abc_2_duration(abc):
  if abc.startswith(u"1/"): abc = abc[1:]
  return _abc_2_duration[abc] #* default_duration // 96


_abc_2_value = {
  "C" : { "C" : 60, "D" : 62, "E" : 64, "F" : 65, "G" : 67, "A" : 69, "B" : 71 },

  "G" : { "C" : 60, "D" : 62, "E" : 64, "F" : 66, "G" : 67, "A" : 69, "B" : 71 },
  "D" : { "C" : 61, "D" : 62, "E" : 64, "F" : 66, "G" : 67, "A" : 69, "B" : 71 },
  "A" : { "C" : 61, "D" : 62, "E" : 64, "F" : 66, "G" : 68, "A" : 69, "B" : 71 },
  "E" : { "C" : 61, "D" : 63, "E" : 64, "F" : 66, "G" : 68, "A" : 69, "B" : 71 },
  "B" : { "C" : 61, "D" : 63, "E" : 64, "F" : 66, "G" : 68, "A" : 70, "B" : 71 },
  "F#": { "C" : 61, "D" : 63, "E" : 65, "F" : 66, "G" : 68, "A" : 70, "B" : 71 },
  "C#": { "C" : 61, "D" : 63, "E" : 65, "F" : 66, "G" : 68, "A" : 70, "B" : 72 },
  
  "F" : { "C" : 60, "D" : 62, "E" : 64, "F" : 65, "G" : 67, "A" : 69, "B" : 70 },
  "Bb": { "C" : 60, "D" : 62, "E" : 63, "F" : 65, "G" : 67, "A" : 69, "B" : 70 },
  "Eb": { "C" : 60, "D" : 62, "E" : 63, "F" : 65, "G" : 67, "A" : 68, "B" : 70 },
  "Ab": { "C" : 60, "D" : 61, "E" : 63, "F" : 65, "G" : 67, "A" : 68, "B" : 70 },
  "Db": { "C" : 60, "D" : 61, "E" : 63, "F" : 65, "G" : 66, "A" : 68, "B" : 70 },
  "Gb": { "C" : 59, "D" : 61, "E" : 63, "F" : 65, "G" : 66, "A" : 68, "B" : 70 },
  "Cb": { "C" : 59, "D" : 61, "E" : 63, "F" : 64, "G" : 66, "A" : 68, "B" : 70 },
  }
def abc_2_value(abc, tonality):
  value = 0
  if abc in "abcdefg":
    abc = abc.upper()
    value += 12
  value += _abc_2_value[tonality][abc]
  return value
  
  
def parse(abc):
  if not isinstance(abc, str): abc = abc.decode("utf8")
  # Simplify bars placed at breakline
  abc = abc.replace("\r", "\n")
  abc = re.sub(u"\\|\\|", u"|", abc)
  abc = re.sub(u"\\|\\s+\\|", u"|", abc)
  abc = re.sub(u":\\|:", u"::", abc)
  
  song      = model.Song()
  partition = model.Partition(song, model.PianoView.default_instrument, view_type = model.PianoView)
  song.add_partition(partition)
  song.mesures = []
  
  titles     = []
  comments   = []
  tempo      = 90
  part_start = -1
  rhythm1    = 4
  rhythm2    = 4
  in_header  = 1
  time       = 0
  syncope    = 0
  note_value_offset = 0
  current_duration = 0
  duration_multiplicator = 1
  in_tripplet   = 0
  ornementation = 0
  in_link       = 0
  start_repeat  = None
  
  def last_mesure_end_time():
    if song.mesures: return song.mesures[-1].end_time()
    return 0
  
  def mesure_ended():
    if song.mesures:
      mesure = song.mesures[-1]
      if (len(song.mesures) == 1) or (mesure is start_repeat):
        #delta_t = mesure.end_time() - partition.notes[-1].end_time()
        delta_t = mesure.end_time() - (time + current_duration)
        if delta_t > 0:
          for note in reversed(partition.notes):
            if note.time < mesure.time: break
            note.time += delta_t
            
  lines = abc.split(u"\n")
  for line in lines:
    if not line: continue
    if "%" in line: line = line.split("%")[0] # ABC Comment
    if (len(line) > 1) and (line[1] == u":") and (line[0] in string.uppercase):
      attr  = line[0]
      value = line[2:]
      if not value: continue
      if value[0] == u" ": value = value[1:]
      
      if   attr == u"T": titles  .append(value)
      elif attr == u"C": song.authors  = value
      elif attr == u"B": comments.append(u"Source: " + value)
      elif attr == u"S": comments.append(u"Source: " + value)
      elif attr == u"A": comments.append(u"Region: " + value)
      elif attr == u"O": comments.append(u"Origin: " + value)
      elif attr == u"H": comments.append(u"History: " + value)
      elif attr == u"N": comments.append(value)
      elif attr == u"I": comments.append(value)
      elif attr == u"L":
        if rhythm2 == 8:
          if   value == u"1"   : default_duration = 384
          elif value == u"1/2" : default_duration = 192
          elif value == u"1/4" : default_duration = 96
          elif value == u"1/8" : default_duration = 48
          elif value == u"1/16": default_duration = 24
          elif value == u"1/32": default_duration = 12
          elif value == u"1/64": default_duration = 6
        else:
          if   value == u"1"   : default_duration = 384
          elif value == u"1/2" : default_duration = 192
          elif value == u"1/4" : default_duration = 96
          elif value == u"1/8" : default_duration = 48
          elif value == u"1/16": default_duration = 24
          elif value == u"1/32": default_duration = 12
          elif value == u"1/64": default_duration = 6
          
        #default_duration = abc_2_duration(value)
        
      elif attr == u"R":
        comments.append(value)
        if (u"jig" in value) or (u"gigue" in value) or (u"syncope" in value) or (u"ternary" in value) or (u"ternaire" in value):
          syncope = 1
        else:
          syncope = 0
      elif attr == u"M":
        if   value == u"C" : rhythm1 = rhythm2 = 4; default_duration = 48
        elif value == u"C|": rhythm1 = rhythm2 = 2; default_duration = 48
        else:
          rhythm1, rhythm2 = [int(i) for i in value.split(u"/")]
          if float(rhythm1) / float(rhythm2) < 0.75: default_duration = 24
          else:                                      default_duration = 48
      elif attr == u"Q":
        if "=" in value:
          duration, value = value.split("=")
          tempo = int(int(value) / abc_2_duration(duration) * 96.0)
        else:
          tempo = int(value)
      elif attr == u"P": pass # XXX
      elif attr == u"K":
        if (len(value) > 1) and (value[1] in u"b#"): partition.tonality = value[:2]
        else:                                        partition.tonality = value[0]
        
    else:
      if in_header:
        in_header = 0
        
      while line:
        print(repr(line))
        
        if   line.startswith(u" "):
          line = line[1:]
        elif line.startswith(u","):
          note.value -= 12
          line = line[1:]
        elif line.startswith(u"'"):
          note.value += 12
          line = line[1:]
        elif line.startswith(u"^"):
          note_value_offset =  1
          line = line[1:]
        elif line.startswith(u"_"):
          note_value_offset = -1
          line = line[1:]
          
        elif line[0] in "abcdefgABCDEFGz":
          if not song.mesures: # first mesure bar is usually missing in ABC
            mesure = model.Mesure(song, 0, tempo, rhythm1, rhythm2, syncope)
            song.add_mesure(mesure)
            start_repeat = mesure
            
          time += current_duration
          duration = int(default_duration * duration_multiplicator)
          if in_tripplet:
            in_tripplet -= 1
            duration = int(duration / 1.5)
            
          if line[0] == u"z": # Pause
            note = None
          else:
            value = abc_2_value(line[0], partition.tonality)
            value += note_value_offset
            note = model.Note(partition, time, duration, value)
            if ornementation: note.volume = 255
            if in_link: note.set_link_fx("link")
            partition.add_note(note)
            
          current_duration = duration
          note_value_offset = 0
          duration_multiplicator = 1
          ornementation = 0
          line = line[1:]
          
        elif line.startswith(u"~"):
          ornementation = 1
          line = line[1:]
          
        elif line.startswith(u"{"): # XXX no grace note in Songwrite 3 yet!
          ornementation = 1
          line = line[line.find(u"}") + 1:]
          
        elif line.startswith(u'"'): # XXX no chord in Songwrite 3 yet!
          line = line[1:]
          line = line[line.find(u'"') + 1:]
          
        elif line.startswith(u"u") or line.startswith(u"v"):
          pass # XXX
          line = line[1:]
          
        elif line.startswith(u"!"):
          pass # XXX
          line = line[1:]
          
        elif line.startswith(u"."):
          ornementation = 1
          line = line[1:]
          
        elif line.startswith(u"-"):
          note.set__link_fx("link")
          line = line[1:]
          
        elif line.startswith(u">"):
          note.duration = current_duration = int(note.duration * 1.5)
          duration_multiplicator = 0.5
          line = line[1:]
        elif line.startswith(u"<"):
          note.duration = current_duration = note.duration // 2
          duration_multiplicator = 1.5
          line = line[1:]
          
        elif line.startswith(u"(3"):
          in_tripplet = 3
          line = line[2:]
          
        elif line.startswith(u"("):
          in_link = 1
          line = line[1:]
          
        elif line.startswith(u")"):
          in_link = 0
          if note: note.set_link_fx(u"")
          line = line[1:]
          
        elif line.startswith(u"|[1") or line.startswith(u"|1"):
          mesure_ended()
          repeated_part = song.mesures.index(start_repeat), len(song.mesures) - 1
          song.playlist.append(model.PlaylistItem(song.playlist, song.mesures.index(start_repeat), len(song.mesures) - 1))
          mesure = model.Mesure(song, last_mesure_end_time(), tempo, rhythm1, rhythm2, syncope)
          song.add_mesure(mesure)
          time = mesure.time
          current_duration = 0
          start_repeat = mesure
          if line.startswith(u"|[1"): line = line[3:]
          else:                       line = line[2:]
          
        elif line.startswith(u":|[2") or line.startswith(u":|2"):
          mesure_ended()
          song.playlist.append(model.PlaylistItem(song.playlist, song.mesures.index(start_repeat), len(song.mesures) - 1))
          song.playlist.append(model.PlaylistItem(song.playlist, *repeated_part))
          mesure = model.Mesure(song, last_mesure_end_time(), tempo, rhythm1, rhythm2, syncope)
          song.add_mesure(mesure)
          time = mesure.time
          current_duration = 0
          start_repeat = mesure
          if line.startswith(u":|[2"): line = line[4:]
          else:                        line = line[3:]
          
        elif line.startswith(u"|:"):
          mesure_ended()
          if start_repeat: song.playlist.append(model.PlaylistItem(song.playlist, song.mesures.index(start_repeat), len(song.mesures) - 1))
          mesure = model.Mesure(song, last_mesure_end_time(), tempo, rhythm1, rhythm2, syncope)
          song.add_mesure(mesure)
          time = mesure.time
          current_duration = 0
          start_repeat = mesure
          line = line[2:]
        elif line.startswith(u":|"):
          mesure_ended()
          song.playlist.append(model.PlaylistItem(song.playlist, song.mesures.index(start_repeat), len(song.mesures) - 1))
          song.playlist.append(model.PlaylistItem(song.playlist, song.mesures.index(start_repeat), len(song.mesures) - 1))
          mesure = model.Mesure(song, last_mesure_end_time(), tempo, rhythm1, rhythm2, syncope)
          song.add_mesure(mesure)
          time = mesure.time
          current_duration = 0
          line = line[2:]
          start_repeat = mesure
        elif line.startswith(u"::"):
          mesure_ended()
          song.playlist.append(model.PlaylistItem(song.playlist, song.mesures.index(start_repeat), len(song.mesures) - 1))
          song.playlist.append(model.PlaylistItem(song.playlist, song.mesures.index(start_repeat), len(song.mesures) - 1))
          mesure = model.Mesure(song, last_mesure_end_time(), tempo, rhythm1, rhythm2, syncope)
          song.add_mesure(mesure)
          time = mesure.time
          current_duration = 0
          start_repeat = mesure
          line = line[2:]
          
        elif line.startswith(u"|") or line.startswith(u"||") or line.startswith(u"[|") or line.startswith(u"|]"):
          mesure_ended()
          mesure = model.Mesure(song, last_mesure_end_time(), tempo, rhythm1, rhythm2, syncope)
          song.add_mesure(mesure)
          time = mesure.time
          current_duration = 0
          if mesure is song.mesures[0]: start_repeat = mesure
          if line.startswith(u"||") or line.startswith(u"[|") or line.startswith(u"|]"): line = line[2:]
          else:                                                                                                                              line = line[1:]
          
        else:
          for abc_duration, duration in _abc_2_duration.items():
            if line.startswith(abc_duration):
              duration = duration * default_duration // 96
              current_duration = duration
              if note: note.duration = duration
              line = line[len(abc_duration):]
              break
            
  song.title    = u" ".join(titles)
  song.comments = u" ".join(comments)
  
  if (partition.notes[-1].time <= song.mesures[-1].time):
    del song.mesures[-1]
    
  if song.playlist.playlist_items:
    try:
      start_repeat_index = song.mesures.index(start_repeat)
      song.playlist.append(model.PlaylistItem(song.playlist, start_repeat_index, len(song.mesures) - 1))
    except ValueError:
      pass
    
    song.playlist.analyse()
    
#   #Fix first bar, since it is often incomplete
#   first_mesure_end_time = song.mesures[0].end_time()
#   first_mesure_notes = [note for note in partition.notes if note.time < first_mesure_end_time]
#   first_mesure_note_end_time = max([note.end_time() for note in first_mesure_notes])
#   if first_mesure_note_end_time < first_mesure_end_time:
#     for note in first_mesure_notes:
#       note.time += first_mesure_end_time - first_mesure_note_end_time

  return song





class TimingError(Exception): pass


class Timer:
  def __init__(self, song):
    self.song      = song
    self.mesures   = song.mesures
    self.time      = 0
    self.mesure_id = 0
    
  def split(self, duration):
    time      = self.time
    mesure_id = self.mesure_id
    durations = []
    while 1:
      mesure_endtime = self.mesures[mesure_id].endtime()
      if time + duration < mesure_endtime:
        dur = duration
        durations.append(dur)
        time += duration
        break
      
      dur = mesure_endtime - time
      durations.append(dur)
      time = mesure_endtime
      mesure_id += 1
      duration -= dur
      if duration == 0: break
      
    return durations
    
  def advance(self, duration, add_rest = 0):
    print("avance de", duration, add_rest)
    
    durations = []
    while 1:
      mesure_endtime = self.mesures[self.mesure_id].endtime()
      if self.time + duration < mesure_endtime:
        dur = duration
        if add_rest: self.add_rest(dur)
        durations.append(dur)
        self.time += duration
        break
      
      dur = mesure_endtime - self.time
      if add_rest: self.add_rest(dur)
      durations.append(dur)
      self.time = mesure_endtime
      self.mesure_id += 1
      duration -= dur
      self.change_mesure(self.mesures[self.mesure_id - 1], self.mesures[self.mesure_id])
      if duration == 0: break
      
    return durations
    
  def goto(self, time):
    if time < self.time: raise TimingError
    self.advance(time - self.time, 1)
    
    
class AbcTimer(Timer):
  def __init__(self, song, abc):
    Timer.__init__(self, song)
    self.abc = abc
    
  def add_rest(self, duration):
    dur = 384
    while duration:
      nb = duration / dur
      duration %= dur
      for i in range(nb): self.abc.write("z" + abcduration(dur))
      dur /= 2

  def start(self):
    for s in self.song.playlist.symbols.get(self.mesures[0], ()):
      if s.startswith(r"\repeat volta"):
        self.abc.write("|:")
        
  def end(self):
    self.abc.write("|]")
    
  def change_mesure(self, left, right):
    print("change mesure")
    #print(left, self.song.playlist.symbols.get(left, None), right, self.song.playlist.symbols.get(right, None))
    
    for s in self.song.playlist.symbols.get(left, ()):
      if   s == r"} % end repeat":                   self.abc.write(":") 
      elif s == r"} % end repeat with alternatives": self.abc.write(":") 
      elif s == r"} % end alternative":              self.abc.write(":")
      elif s == r"} % end last alternative":         self.abc.write(":")
      
    self.abc.write("|")
    
    for s in self.song.playlist.symbols.get(right, ()):
      if   s.startswith(r"\repeat volta"):         self.abc.write(":")
      elif s.startswith(r"{ % start alternative"): self.abc.write("[" + s.split()[-1])
      
      
      
      
def abctab(s):
  abc = cStringIO.StringIO()
  print("%%!abctab2ps -paper %s" % {"a3paper" : "a3", "a4paper" : "a4", "a5paper" : "a5", "letterpaper" : "letter", }[globdef.config.PAGE_FORMAT], file = abc)
  print("X:1", file = abc)
  print("L:1/4", file = abc)
  if s.title   : print("T:%s" % s.title, file = abc)
  if s.authors : print("C:%s" % s.authors, file = abc)
  if s.comments: print("N:%s" % s.comments.replace("\n", "N:\n"), file = abc)
  if s.filename: print("F:%s" % s.filename, file = abc)
  
  for partition in s.partitions:
    if hasattr(partition, "tonality"):
      print("K:%s" % partition.tonality, file = abc)
      break
  else:
    print("K:C", file = abc)
    
  nb = 0
  for partition in s.partitions:
    if   partition.view.__class__.__name__ == "Tablature":
      clef = " clef=guitartab"
      tab  = 6
    elif partition.view.__class__.__name__ == "Staff":
      clef = ""
      tab  = 0
    else: continue
    
    lyric    = None # XXX
    tonality = getattr(partition, "tonality", "C")
    
    print("%", file = abc)
    print("V:%s name=%s%s" % (nb, partition.header, clef), file = abc)
    print("K:%s" % tonality, file = abc)
    
    nb += 1
    
    timer = AbcTimer(s, abc)
    timer.start()
    
    chords = chordify(partition.notes)
    for chord in chords:
      timer.goto(chord.time)
      
      if tab:
        for duration in timer.split(chord.duration):
          frets = [","] * tab
          for note in chord.notes:
            stringid = partition.view.stringid(note)
            frets[stringid] = _tabnotes[note.value - partition.view.strings[stringid].basenote]

          while frets and (frets[-1] == ","): frets.pop()
          abc.write("[%s%s]" % ("".join(frets), abcduration(duration)))
          
          timer.advance(duration)
          
      else:
        for duration in timer.advance(chord.duration):
          for note in chord.notes:
            abc.write(
              "[" * ((len(chord.notes) > 1) and (note is chord.notes[0])) +
              _notes[tonality][note.value % 12] +
              "," * (4 - note.value / 12) +
              "'" * (note.value / 12 - 4) +
              abcduration(duration, note) +
              "]" * ((len(chord.notes) > 1) and (note is chord.notes[-1]))
              )
      
    timer.end()
    
  #open("/home/jiba/tmp/test.abc", "w").write(abc.getvalue())
  return abc.getvalue()



class Chord:
  def __init__(self, time):
    self.time  = time
    self.notes = []
    self.duration = 1000000
    
  def add(self, note):
    self.notes.append(note)
    self.duration = min(note.duration, self.duration)
    
  def __repr__(self):
    return "<Chord at %s duration %s %s>" % (self.time, self.duration, self.notes)
    
def chordify(notes):
  chords   = []
  previous = None
  notes    = notes[:]
  
  while notes:
    chord = Chord(notes[0].time)
    chords.append(chord)
    
    while notes:
      if notes[0].time == chord.time: chord.add(notes.pop(0))
      else: break
      
    if previous:
      if previous.time + previous.duration > chord.time: # Overlaps !!!
        previous.duration = chord.time - previous.time
        
    previous = chord

  return chords
  














# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2007 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import songwrite3.plugins as plugins


class ABCImportPlugin(plugins.ImportPlugin):
  def __init__(self):
    plugins.ImportPlugin.__init__(self, "ABC", [".abc", ".txt"])
    
  def import_from_string(self, data):
    import songwrite3.plugins.abc.abc
    return songwrite3.plugins.abc.abc.parse(data)
    
ABCImportPlugin()





# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2001-2007 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, array, re
import string as string_module # not a guitar string !
from io import StringIO

import songwrite3.model as model


def parse(texttab):
  ASCII_STRING    = "-="
  ASCII_BAR       = "|"
  ASCII_STRESSED  = "'"
  ASCII_HAMMER    = "hpl"
  ASCII_SLIDE     = "/\\s"
  ASCII_BEND      = "b"
  ASCII_DEAD_NOTE = "xd"
  ASCII_TREMOLO   = "~vt"
  ASCII_TAB       = ASCII_STRING + ASCII_BAR + string_module.digits + ASCII_STRESSED + ASCII_HAMMER + ASCII_SLIDE + ASCII_BEND + ASCII_DEAD_NOTE + ASCII_TREMOLO
  
  TITRE_AUTHOR = re.compile(r".*\(.*\)")
  COPYRIGHT    = re.compile(r"(C|c)opy(right|left)\s*[0-9]{2,4}")
  TEMPO        = re.compile(r"((?<=(T|t)empo(:|=)\s)|(?<=(T|t)empo(:|=))|(?<=(T|t)empo\s(:|=)\s)|(?<=(T|t)empo\s))([0-9]+)")
  
  song = model.Song()
  lines = texttab.split("\n")
  
  tempo = TEMPO.findall(texttab)
  if tempo: tempo = tempo[0][-1]
  else:     tempo = 60
  
  i = 0
  header = ""
  song_comments_ended = 0
  if (not lines[i]) or (not lines[i][0] in ASCII_TAB): # First line is guessed to be the title.
    if TITRE_AUTHOR.match(lines[i]):
      start = lines[i].find("(")
      song.title   = lines[i][: start]
      song.authors = lines[i][start + 1 : lines[i].find(")")]
    else: song.title = lines[i]
    i = i + 1
    
  while 1:
    if   not lines[i]: song_comments_ended = 1
    elif not lines[i][0] in ASCII_TAB:
      if COPYRIGHT.search(lines[i]): song.copyright = lines[i]
      else:
        if song_comments_ended:
          if header: header = header + "\n" + lines[i]
          else:      header = lines[i]
        else:
          if song.comments: song.comments = song.comments + "\n" + lines[i]
          else:              song.comments = lines[i]
    else: break
    i = i + 1
    
  # Compute rythm
  if lines[i][0] in ASCII_BAR: mesure_len = lines[i][1:].find("|")
  else:                        mesure_len = lines[i].find("|")
  
  if   mesure_len % 4 == 0: unit = float(mesure_len / 4) # Guess it is 4/4
  elif mesure_len % 3 == 0: unit = float(mesure_len / 3) # Guess it is 3/4
  
  pos              = []
  last_bar_pos     = 0
  partition        = model.Partition(song)
  partition.header = header
  # adrian@sixfingeredman.net: hack to guess tuning
  if "DADGAD" in header: partition.view = model.GuitarDADGADView(partition)
  else:                  partition.view = model.GuitarView      (partition)
  current_note     = None
  next_note        = model.Note(partition, 0, 96, 0)
  song.partitions.append(partition)
  song.mesures *= 0
  
  while (i < len(lines)):
    string_id = 0
    
    while (i < len(lines)) and lines[i] and (lines[i][0] in ASCII_TAB):
      j = 0
      if string_id >= len(pos) - 1: pos.append(0)
      
      while j < len(lines[i]):
        char = lines[i][j]
        
        if char in ASCII_BAR:
          if (string_id == 0) and (partition is song.partitions[0]):
            if j > 1:
              song.add_mesure(model.Mesure(song,
                                           int(round(last_bar_pos / unit * 96.0)),
                                           rythm1 = int(round((pos[string_id] - last_bar_pos) / unit)),
                                           tempo  = tempo))
            last_bar_pos = pos[string_id]
          j = j + 1
          continue
        
        elif char in string_module.digits:
          if (j + 1 < len(lines[i])) and (lines[i][j + 1] in string_module.digits):
            fret = int(char + lines[i][j + 1])
            pos[string_id] = pos[string_id] + 1
            j = j + 1
          else: fret = int(char)
          next_note.value     = partition.view.strings[string_id].base_note + fret
          next_note.time      = int(round((pos[string_id] - 1)/ unit * 96.0))
          next_note.string_id = string_id
          partition.notes.append(next_note)
          
          current_note = next_note
          next_note = model.Note(partition, 0, 96, 0)
          
          
          
        elif char in ASCII_STRESSED : next_note.volume = 255
        elif char in ASCII_TREMOLO  : next_note.fx = "tremolo"
        elif char in ASCII_DEAD_NOTE: next_note.fx = "dead"
        elif char in ASCII_BEND     : next_note.fx = "bend"
        elif char in ASCII_HAMMER:
          current_note.link_fx = "link"
          current_note.link_to(next_note)
        elif char in ASCII_SLIDE:
          current_note.link_fx = "slide"
          current_note.link_to(next_note)
        elif char in ASCII_STRING: pass
        else: print("* parse_texttab * Unknown character %s at line %s, position %s ! Skipping." % (char, i, j), file = sys.stderr)
        
        pos[string_id] = pos[string_id] + 1
        j = j + 1
        
      i = i + 1
      string_id = string_id + 1
      
    while (i < len(lines)) and (not lines[i]): i = i + 1
    
    if (i < len(lines)) and ((len(lines[i]) < 3) or (not lines[i][0] in ASCII_TAB) or (not lines[i][1] in ASCII_TAB)): break
    
    
  # Compute notes' durations
  partition.notes.sort()
  last_notes = []
  for note in partition.notes:
    if last_notes and (last_notes[0].time != note.time):
      for last_note in last_notes:
        last_note.duration = note.time - last_note.time
      last_notes = []
    last_notes.append(note)
    
  last_duration = song.mesure_at(last_notes[0].time, 1).end_time() - last_notes[0].time
  for last_note in last_notes:
    last_note.duration = last_duration
    
  return song


def texttab(song, cut_tab_at = 80):
  "texttab(song) -> string -- Generates an ascii tablature from a song"
  min_duration = sys.maxsize
  ternary      = 0
  partitions   = []
  
  for partition in song.partitions:
    if isinstance(partition, model.Partition) and isinstance(partition.view, model.TablatureView):
      partitions.append(partition)
      
      for note in partition.notes:
        dur = note.duration
        if note.is_triplet(): # Triplets are a hack...
          ternary = 1
          if note.duration == 32: dur = 48
          if note.duration == 16: dur = 24
          if note.duration ==  8: dur = 12
        else:
          rest = note.time % min_duration
          if rest > 0: min_duration = rest
          
        if dur < min_duration: min_duration = dur
        
  min_duration = float(min_duration)
  
  slotsize = 2
  for partition in song.partitions:
    if isinstance(partition, model.Partition) and isinstance(partition.view, model.TablatureView):
      for note in partition.notes:
        if note.is_triplet():
          slotsize = 3
          break
        
  duration = min_duration / slotsize
  
  alltabs = StringIO()
  
  alltabs.write("%s (%s)\n" % (song.title, song.authors))
  if song.copyright: alltabs.write(song.copyright + "\n")
  if song.comments : alltabs.write(song.comments  + "\n")

  alltabs.write("\n")
  
  for partition in song.partitions:
    if   isinstance(partition, model.Lyrics):
      alltabs.write(partition.header + "\n")
      text = partition.text.replace("\t_", "").replace("_", "").replace("-\t", "").replace("\t", " ").replace("\n", "").replace("\\\\", "\n").replace("  ", " ")
      alltabs.write("\n".join(i.strip() for i in text.split("\n")))
      alltabs.write("\n")
      
    elif isinstance(partition.view, model.TablatureView):
      alltabs.write(partition.header + "\n")
      strings = partition.view.strings
      tabs    = [array.array("u", "-" * int(round(partition.end_time() / min_duration * slotsize))) for string in strings]
      
      def string_id(note):
        
        return getattr(note, "string_id", len(strings) - 1)
      
      for note in partition.notes:
        sid  = string_id(note)
        time = int(round(note.time / min_duration * slotsize))
        text = strings[sid].value_2_text(note)
        
        if note.linked_from:
          if   note.linked_from.link_fx == "link":
            link_type = note.linked_from.link_type()
            if   link_type ==  0: continue # Don't show linked note, since ascii tab doesn't matter note duration !
            elif link_type == -1: tabs[sid][time] = "p"
            elif link_type ==  1: tabs[sid][time] = "h"
            
          elif note.linked_from.link_fx == "slide":
            if note.linked_from.value > note.value: tabs[sid][time] = "\\"
            else:                                   tabs[sid][time] = "/"
              
        if   note.fx == "dead"   : tabs[sid][time] = "x"
        elif note.fx == "tremolo": tabs[sid][time] = "~"
        elif note.fx == "bend"   : tabs[sid][time] = "b"
        
        if (note.volume > 220) and (tabs[sid][time] == "-"): tabs[sid][time] = "'" # Stressed note
        
        if len(text) == 1: tabs[sid][time + 1] = text
        else:
          tabs[sid][time + 1] = text[0]
          tabs[sid][time + 2] = text[1]
          
      tabs = [tab.tounicode() for tab in tabs] # Turn arrays to strings.
      
      mesure = song.mesures[0]
      i = 0
      while i < len(song.mesures):
        mesures = []
        length = 1
        while i < len(song.mesures):
          length = length + mesure.duration / min_duration * slotsize + 1
          if length > cut_tab_at: break
          mesures.append(mesure)
          
          i = i + 1
          if i >= len(song.mesures): break
          mesure = song.mesures[i]
          

        string_base_note_length  = max([len(model.note_label(string.base_note, False)) for string in strings])
        string_base_note_pattern = "%%-%ss" % string_base_note_length
        
        for sid in range(len(tabs)):
          tab = tabs[sid]
          
          # adrian@sixfingeredman.net: this screws up the importer
          # alltabs.write(string_base_note_pattern % model.note_label(strings[sid].base_note, False))
          
          for mes in mesures:
            content = tab[int(round(mes.time / min_duration * slotsize)) : int(round(mes.end_time() / min_duration * slotsize))]
            if not content:
              i = sys.maxsize
              break
            alltabs.write("|" + content)
          alltabs.write("|\n")
          
        alltabs.write("\n")
        
  return alltabs.getvalue()





# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2001-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, tempfile, shutil
from io import StringIO

import songwrite3.globdef  as globdef
import songwrite3.model    as model
import songwrite3.plugins  as plugins

import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtGui     as qtgui

def escape_latex(s):
  return s.replace("°", "\\symbol{6}").replace("#", "\\#").replace("_", "\\_").replace("~", "\\~").replace("&", "\\&").replace("%", "\\%").replace("$", "\\$").replace("^", "\\^")

lang_iso2latex = { # Languages supported by LaTeX with Babel
  "" : "\\usepackage[]{babel}",
  }
lang_iso2latex = { # Languages supported by LaTeX with Babel
  "ar" : "arabic",
  "hy" : "armenian",
  "eu" : "basque",
  "bg" : "bulgarian",
  "ca" : "catalan",
  "hr" : "croatian",
  "cs" : "czech",
  "da" : "danish",
  "nl" : "dutch",
  "en" : "english",
  "eo" : "esperanto",
  "et" : "estonian",
  "fa" : "farsi",
  "fi" : "finnish",
  "fr" : "french",
  "gl" : "galician",
  "de" : "german",
  "el" : "greek",
  "hu" : "hungarian",
  "is" : "icelandic",
  "as" : "assamese",
  "bn" : "bengali",
  "gu" : "gujarati",
  "hi" : "hindi",
  "kn" : "kannada",
  "ml" : "malayalam",
  "mr" : "marathi",
  "or" : "oriya",
  "pa" : "panjabi",
  "ta" : "tamil",
  "te" : "telugu",
  "id" : "indonesian",
  "ia" : "interlingua",
  "ga" : "irish",
  "it" : "italian",
  "ku" : "kurmanji",
  "lo" : "lao",
  "la" : "latin",
  "lv" : "latvian",
  "lt" : "lithuanian",
  "mn" : "mongolian",
  "nb" : "bokmal",
  "nn" : "nynorsk",
  "pl" : "polish",
  "pt" : "portuguese",
  "ro" : "romanian",
  "ru" : "russian",
  "sa" : "sanskrit",
  "sr" : "serbian",
  "sk" : "slovak",
  "sl" : "slovenian",
  "es" : "spanish",
  "sv" : "swedish",
  "tr" : "turkish",
  "tk" : "turkmen",
  "uk" : "ukrainian",
  "cy" : "welsh",
  "--" : "",
  }

def latexify_songbook(tmp_dir, songbook):
  latexes = []
  lang    = None
  
  for song_ref in songbook.song_refs:
    song = song_ref.get_song()
    latexes.append(latexify_song(tmp_dir, song, 1))
    if not lang: lang = song.lang

  lang_latex = lang_iso2latex[lang]
  babel_code = "\\usepackage[%s]{babel}" % lang_latex
  open(os.path.join(tmp_dir, "main.latex"), "w").write(("""
\\documentclass[%s,twoside,10pt]{article}
\\usepackage[T1]{fontenc}
\\usepackage[utf8]{inputenc}
\\usepackage[lmargin=%scm,rmargin=%scm,tmargin=%scm,bmargin=%scm]{geometry}
\\usepackage{graphicx}
\\usepackage{lmodern,aecompl}
\\usepackage{multicol} 
%s

\\begin{document}

\\sloppy

\\title {%s}
\\author{%s}
\\date  {}
\\maketitle
\\vfill
%s
\\begin{multicols}{2}
\\tableofcontents
\\end{multicols}
\\pagebreak

%s


\\end{document}
""" % (globdef.config.PAGE_FORMAT,
       globdef.config.PAGE_MARGIN_INTERIOR,
       globdef.config.PAGE_MARGIN_EXTERIOR,
       globdef.config.PAGE_MARGIN_TOP,
       globdef.config.PAGE_MARGIN_BOTTOM,
       babel_code,
       escape_latex(songbook.title),
       escape_latex(songbook.authors),
       escape_latex(songbook.comments),
       """
\\newpage
""".join(latexes),
       )))


PDF_ID = 0
def latexify_song(tmp_dir, song, songbook = 0):
  global PDF_ID
  from songwrite3.plugins.pdf.pdf_canvas import BaseCanvas, PDFCanvas
  
  song.remove_unused_mesures()
  
  latex = StringIO()
  
  partitions = song.partitions
  
  lang = lang_iso2latex.get(song.lang, None)
  if lang: babel_code = "\\usepackage[%s]{babel}" % lang
  else:    babel_code = ""
  
  if songbook:
    latex.write("""
\\addcontentsline{toc}{subsection}{%s}

""" % escape_latex(song.title))
    
  else:
    latex.write("""
\\documentclass[%s,twoside,10pt]{article}
\\usepackage[T1]{fontenc}
\\usepackage[utf8]{inputenc}
\\usepackage[lmargin=%scm,rmargin=%scm,tmargin=%scm,bmargin=%scm,headheight=0cm,headsep=0cm]{geometry}
\\usepackage{graphicx}
\\usepackage{aecompl}""" % (globdef.config.PAGE_FORMAT,
       globdef.config.PAGE_MARGIN_INTERIOR,
       globdef.config.PAGE_MARGIN_EXTERIOR,
       globdef.config.PAGE_MARGIN_TOP,
       globdef.config.PAGE_MARGIN_BOTTOM,
       ))
    if song.print_lyrics_columns > 1: latex.write("""\\usepackage{multicol}""")
    latex.write("""
%s
\\setlength{\\topskip}{0pt}

\\begin{document}

\\sloppy

""" % babel_code)

  if song.title:
    latex.write("""
\\begin{center}\\begin{LARGE} %s \\end{LARGE}\\end{center}
""" % escape_latex(song.title))
    
  date = song.date()
  if song.authors or date:
    if   not date:         authors_date = song.authors
    elif not song.authors: authors_date = date
    else:                  authors_date = "%s (%s)" % (song.authors, date)
    latex.write("""
\\begin{center}\\begin{large} %s \\end{large}\\end{center}
""" % escape_latex(authors_date))
    
  if song.comments:
    latex.write("""
%s
""" % escape_latex(song.comments).replace("\\n", "\\n\\n"))

  max_time = max([10] + [partition.end_time() for partition in song.partitions if isinstance(partition, model.Partition)])
  max_width = {
    "a3paper"    : 29.7,
    "a4paper"    : 21.0,
    "a5paper"    : 14.8,
    "letterpaper": 21.6,
    "legalpaper" : 21.6,
    }[globdef.config.PAGE_FORMAT] - (globdef.config.PAGE_MARGIN_INTERIOR + globdef.config.PAGE_MARGIN_EXTERIOR)
  max_width = int(round(max_width * 56.8421052631579))
  
  max_mesures_width = 0
  mesuress = []
  selected_mesures = []
  canvas = PDFCanvas(song)
  #canvas.set_default_font_size(song.printfontsize)
  canvas.update_mesure_size()
  width = canvas.start_x
  
  zoom = (max_width - canvas.start_x - 1) / float(song.mesures[min(song.print_nb_mesures_per_line, len(song.mesures)) - 1].end_time() )
  for mesure in song.mesures:
    if len(selected_mesures) >= song.print_nb_mesures_per_line:
      mesuress.append(selected_mesures)
      selected_mesures = []
    selected_mesures.append(mesure)
  if selected_mesures:
    if mesuress and (len(selected_mesures) == 1) and (song.print_nb_mesures_per_line > 3):
      mesuress[-1].append(selected_mesures[0])
    else:
      mesuress.append(selected_mesures)
      
  def mesure_2_repeat(mesure):
    start_repeat = end_repeat = 0
    for symbol in song.playlist.symbols.get(mesure) or []:
      if   symbol.startswith(r"\repeat"):    start_repeat = 1
      elif symbol == r"} % end alternative": end_repeat   = 1
      elif symbol == r"} % end repeat":      end_repeat   = 1
    return start_repeat, end_repeat
    
  def mesures_2_x_width_zoom(mesures, is_last, previous_mesures, previous_zoom):
    if mesures[0] is song.mesures[0]:
      x   = 0
      dx  = 0
      start_repeat, end_repeat = mesure_2_repeat(mesures[0])
      if start_repeat:                  x +=0.3 * canvas.scale; dx -=0.3 * canvas.scale
    else:
      x   = canvas.time_2_x(mesures[0].time) - canvas.start_x
      dx  = x  - mesures[ 0].time
      
      start_repeat, end_repeat = mesure_2_repeat(mesures[0])
      if   start_repeat and end_repeat: x -= 17 * canvas.scale; dx -= 17 * canvas.scale
      elif start_repeat:                x -= 16 * canvas.scale; dx -= 16 * canvas.scale
      elif end_repeat:                  x -=  2 * canvas.scale; dx -=  2 * canvas.scale
      else:                             x -=  2 * canvas.scale; dx -=  2 * canvas.scale
      
    x2    = canvas.time_2_x(mesures[-1].end_time()) - canvas.start_x
    dx2   = x2 - mesures[-1].end_time()
    
    try:    next_mesure = song.mesures[mesures[-1].get_number() + 1]
    except: next_mesure = None
    start_repeat, end_repeat = mesure_2_repeat(next_mesure)
    if   start_repeat and end_repeat: x2 -= 15 * canvas.scale; dx2 -= 15 * canvas.scale
    elif start_repeat:                x2 -= 16 * canvas.scale; dx2 -= 16 * canvas.scale
    elif end_repeat:                  x2 -=  2 * canvas.scale; dx2 -=  2 * canvas.scale
    else:                             x2 -=1.1 * canvas.scale; dx2 -=1.1 * canvas.scale
    
    if is_last and not (len(mesures) < len(previous_mesures)):
      x2 +=  5 * canvas.scale; dx2 +=  5 * canvas.scale
      
    width = x2 - x
    
    if is_last and (len(mesures) < len(previous_mesures)):
      zoom = previous_zoom
      width = (dx2 - dx) + (width - (dx2 - dx)) * zoom + canvas.start_x
      if not end_repeat: width += 5 * canvas.scale
    else:
      zoom  = (max_width - canvas.start_x - (dx2 - dx)) / float(width - (dx2 - dx))
      width = max_width
      
    x = dx + (x - dx) * zoom
    return x, width, zoom

  # We need a second canvas; the first one is only for computation
  pdf_canvas = PDFCanvas(song)
  
  latex.write("""
\\begin{flushleft}
""")
  previous_mesures = []
  previous_zoom    = 0.0
  for mesures in mesuress:
    x, width, zoom = mesures_2_x_width_zoom(mesures, mesures is mesuress[-1], previous_mesures, previous_zoom)
    pdf_filename = os.path.join(tmp_dir, "pdf_file_%s.pdf" % PDF_ID)
    pdf_canvas.render(pdf_filename, x, width, mesures[0].time, mesures[-1].end_time(), zoom)
    previous_mesures = mesures
    previous_zoom    = zoom

    if hasattr(qtgui, "QPageLayout"): # setResolution is used for changing resolution.
      latex.write("""
      \\noindent\\includegraphics[clip,scale=0.5]{pdf_file_%s.pdf}\\par""" % PDF_ID)
    else:
      latex.write("""
      \\noindent\\includegraphics[clip,scale=8.3333]{pdf_file_%s.pdf}\\par""" % PDF_ID)
    if mesures is mesuress[-1]: latex.write("""\n""")
    else:                       latex.write("""\\bigskip\n""")
    PDF_ID += 1
    
  latex.write("""
\\end{flushleft}
""")
  latex.write("""
\\vfill
""")
  if song.print_lyrics_columns > 1:
    latex.write("""
\\begin{multicols}{%s}
""" % song.print_lyrics_columns)
  

  first = 1
  for partition in partitions:
    if isinstance(partition, model.Lyrics):
      if not partition.show_all_lines_after_score: continue
      if first:
        first = 0
        #latex.write("""\\noindent """)
      else:
        #if song.print_lyrics_columns > 1:
          latex.write("""
\\smallskip{}

""")
        #else:
        #  latex.write("""\\\\
#\\noindent """)
      if partition.header:
        latex.write("""\\textbf{%s}\\nopagebreak[4]\\\\%%
""" % escape_latex(partition.header))
      else:
        latex.write("""\\noindent """)
        


      text = partition.text.replace("_-\t", "").replace("_\t", " ").replace("-\t", "").replace("\t", " ").replace("_", "").replace(" ,", ",").replace(" ,", ",")
      if text.endswith("\\"): text = text[:-2] # Remove last breakline
      latex.write(escape_latex(text))
      
  if song.print_lyrics_columns > 1:
    latex.write("""
\end{multicols} 
""")
#  else:
#    latex.write("""
#\\vfill
#""")
    
    
  if not songbook:
    latex.write("""
\\end{document}
""")

  if songbook:
    return latex.getvalue()
  else:
    open(os.path.join(tmp_dir, "main.latex"), "w").write(latex.getvalue())


PDF_PAGE_FORMATS = {
  "a3paper"     : "a3",
  "a4paper"     : "a4",
  "a5paper"     : "a5",
  "letterpaper" : "letter",
  "legalpaper"  : "legal",
  }

def do(command):
  print(file = sys.stderr)
  print("Running '%s'..." % command, file = sys.stderr)
  os.system(command)
  
def pdfy(song):
  tmp_dir = tempfile.mkdtemp()
  print(file = sys.stderr)
  print("Using temporary directory '%s'..." % tmp_dir, file = sys.stderr)
  
  if isinstance(song, model.Songbook):
    latexify_songbook(tmp_dir, song)
  else:
    latexify_song    (tmp_dir, song)
    
  do("cd %s; pdflatex -interaction=nonstopmode main.latex" % tmp_dir)
  do("cd %s; pdflatex -interaction=nonstopmode main.latex" % tmp_dir) # For table of content
  
  with open(os.path.join(tmp_dir, "main.pdf"), "rb") as f: r = f.read()
  shutil.rmtree(tmp_dir)
  return r




# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2008 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtGui     as qtgui

import editobj3

import songwrite3.canvas as canvas


class FingeringDrawer(canvas.PartitionDrawer):
  def __init__(self, canvas_, partition, compact, String):
    canvas.PartitionDrawer.__init__(self, canvas_, partition, compact)
    self.height = 300
    self.strings = [String(self)]
    self.string_height = self.canvas.default_line_height * 0.3
    for hole in next(iter(partition.view.fingerings.values())):
      if hole is None: self.string_height += self.canvas.default_line_height * 0.3
      else:            self.string_height += self.canvas.default_line_height * 0.6
    self.stem_extra_height = self.canvas.default_line_height
    
    self.lh2  = self.canvas.default_line_height / 2
    self.lh3  = self.canvas.default_line_height / 3
    self.lh4  = self.canvas.default_line_height / 4.5
    self.lh7  = self.canvas.default_line_height / 7
    self.lh_6 = self.canvas.default_line_height * 0.6

  def partition_listener(self, partition, type, new, old):
    if type is object:
      if (new.get("instrument_tonality") != old.get("instrument_tonality")):
        self.canvas.render_all()
        
  def note_width(self, note): return 9.0 * self.scale
  
  def note_string_id(self, note): return 0 # No "string" notion in fingering notation!
  def y_2_string_id(self, y, force = 0):
    if force:
      if   self.y + self.start_y > y: return -1
      elif y > self.y + self.height:  return  1
      return 0
    else:
      if self.y + self.start_y <= y <= self.y + self.height: return 0
      return None
  def string_id_2_y(self, string_id = 0):
    return self.y + self.start_y + self.stem_extra_height + self.string_height // 2
  def draw_mesure(self, ctx, time, mesure, y, height):
    canvas.PartitionDrawer.draw_mesure(self, ctx, time, mesure, self.y + self.start_y + self.stem_extra_height, self.string_height)
  def draw_strings(self, ctx, x, y, width, height):
    self.height = int(self.start_y + self.stem_extra_height + self.string_height + self.stem_extra_height_bottom + 1)
    
  def note_text(self, note): return "_"
  
  def note_dimensions(self, note):
    x = self.canvas.time_2_x(note.time)
    y = self.string_id_2_y()
    return (
      x + 3.0 * self.scale,
      y - self.string_height // 2,
      x + max(note.duration * self.canvas.zoom, self.canvas.char_h_size * 1.9),
      y + self.string_height // 2,
      )
    
  def draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure = None, previous_time = -32000, next_time = 32000):
    x = x + self.canvas.default_line_height // 2
    canvas.PartitionDrawer.draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure, previous_time, next_time)
    
  def draw_note(self, ctx, note, string_id, y):
    color = self.note_color(note)
    ctx.setPen(color)
    ctx.setBrush(color)
    
    x = self.canvas.time_2_x(note.time)
    y -= self.string_height // 2
    
    if note.is_dotted():
      #ctx.arc(x + self.canvas.default_line_height // 2 + 4 * self.scale, y - 3 * self.scale, 1.5, 0, 6.2831853071795862)
      #ctx.fill()
      ctx.drawEllipse(
        x + self.canvas.default_line_height // 2 + 2.5 * self.scale,
        y - 4.5 * self.scale,
        1.5 * self.scale,
        1.5 * self.scale,
      )
      
    fingering = self.partition.view.note_2_fingering(note)
    if fingering:
      holes  = fingering[:-1]
      breath = fingering[-1]
    else:
      holes = []
      breath = "?"

    for hole in holes:
      if   hole is None:
        y += self.lh3
        continue
      elif hole == 1:
        ctx.drawEllipse(x + self.lh2 - self.lh4, y + self.lh3 - self.lh4, self.lh4 * 2.0 + 0.5, self.lh4 * 2.0 + 0.5)
      elif hole == 0:
        ctx.setBrush(qtcore.Qt.transparent)
        ctx.drawEllipse(x + self.lh2 - self.lh4, y + self.lh3 - self.lh4, self.lh4 * 2.0 + 0.5, self.lh4 * 2.0 + 0.5)
        ctx.setBrush(color)
      elif hole == 0.5:
        ctx.setBrush(qtcore.Qt.transparent)
        ctx.drawEllipse(x + self.lh2 - self.lh4, y + self.lh3 - self.lh4, self.lh4 * 2.0 + 0.5, self.lh4 * 2.0 + 0.5)
        ctx.setBrush(color)
        ctx.drawChord  (x + self.lh2 - self.lh4, y + self.lh3 - self.lh4, self.lh4 * 2.0 + 0.5, self.lh4 * 2.0 + 0.5, 0.0, 180 * 16)
        
      else:
        x2 = x + self.lh2 - self.lh3 + self.lh7
        for sub_hole in hole:
          if   sub_hole == 1:
            ctx.drawEllipse(x2 - self.lh7, y + self.lh3 - self.lh7, self.lh7 * 2.0, self.lh7 * 2.0)
          elif sub_hole == 0:
            ctx.setBrush(qtcore.Qt.transparent)
            ctx.drawEllipse(x2 - self.lh7, y + self.lh3 - self.lh7, self.lh7 * 2.0, self.lh7 * 2.0)
            ctx.setBrush(color)
          x2 += self.lh3
      y += self.lh_6


    if breath:
      ctx.drawText(x + self.canvas.default_line_height * 0.2,
                   y + self.canvas.default_line_height * 0.8,
                   breath)
      
    if note.fx     : getattr(self, "draw_note_fx_%s" % note.fx     )(ctx, note, string_id)
    if note.link_fx: getattr(self, "draw_note_fx_%s" % note.link_fx)(ctx, note, string_id)
    ctx.setPen(qtcore.Qt.black)
    ctx.setBrush(qtcore.Qt.black)
    

  def on_touchscreen_new_note(self, event):
    y = event.y() - self.y - self.start_y - self.stem_extra_height
    hole_clicked = -1
    for hole in next(iter(self.partition.view.fingerings.values())):
      if y < 0: break
      hole_clicked += 1
      if hole is None: y -= self.canvas.default_line_height * 0.3
      else:            y -= self.canvas.default_line_height * 0.6
    new_value = self.strings[0].on_click_at(hole_clicked)
    self.canvas.on_number_typed(new_value)
    return 1 # OK to change note by dragging up or down

    
class SingleString(object):
  def __init__(self, drawer, notation_pos = -1):
    self.drawer       = drawer
    self.notation_pos = notation_pos
    
  def get_base_note(self): return self.drawer.partition.view.get_bell_note()
  base_note = property(get_base_note)
  
  def text_2_value(self, note, text): return self.base_note
  def value_2_text(self, note): return "o"
  
  def __str__(self): return "Single string"
  
  def width(self): return -1
  
  def on_click_at(self, hole): return "0"
  

class TinWhistleSingleString(SingleString):
  def text_2_value(self, note, text):
    if   text ==  "0": return self.base_note + 11
    elif text ==  "9": return self.base_note + 10
    elif text ==  "1": return self.base_note + 9
    elif text ==  "2": return self.base_note + 7
    elif text ==  "3": return self.base_note + 5
    elif text ==  "4": return self.base_note + 4
    elif text ==  "5": return self.base_note + 2
    elif text ==  "6": return self.base_note
    elif text == "00": return self.base_note + 23
    elif text == "11": return self.base_note + 21
    elif text == "22": return self.base_note + 19
    elif text == "33": return self.base_note + 17
    elif text == "44": return self.base_note + 16
    elif text == "55": return self.base_note + 14
    elif text == "66": return self.base_note + 12
    if len(text) > 2: return self.text_2_value(note, text[-2:])
    if len(text) > 1: return self.text_2_value(note, text[-1])
    return self.base_note
  
  def on_click_at(self, note, hole):
    current_fingering = self.drawer.partition.view.note_2_fingering(note)
    g8 =   0
    if current_fingering and (current_fingering[-1] == "+"): breath = 12
    else:                                                     breath = 0
    if   hole <  0: return self.base_note + g8 + breath + 11
    elif hole == 0: return self.base_note + g8 + breath + 9
    elif hole == 1: return self.base_note + g8 + breath + 7
    elif hole == 2: return self.base_note + g8 + breath + 5
    #    hole == 3 is the "None" separator
    elif hole == 4: return self.base_note + g8 + breath + 4
    elif hole == 5: return self.base_note + g8 + breath + 2
    elif hole == 6: return self.base_note + g8 + breath
    elif hole == 7: # Toggle breath
      if breath: return note.value - 12
      else:      return note.value + 12
    

  def on_click_at(self, hole):
    if   hole <  0: return "0"
    elif hole == 0: return "1"
    elif hole == 1: return "2"
    elif hole == 2: return "3"
    #    hole == 3 is the "None" separator
    elif hole == 4: return "4"
    elif hole == 5: return "5"
    elif hole == 6: return "6"    



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2009-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import songwrite3.model   as model
import songwrite3.plugins as plugins

class LyreView(model.TablatureView):
  view_class               = "lyre_tablature"
  default_instrument       = 46
  default_icon_filename    = "lyre.png"
  link_type                = model.LINK_NOTES_DEFAULT
  automatic_string_id      = 1
  use_harmonics_for_octavo = 1
  
  def __init__(self, partition, name = "", strings = None):
    if partition:
      if not hasattr(partition, "let_ring"): partition.let_ring = 1
    model.TablatureView.__init__(self, partition, name, strings)
    self.tonality = "C"
    self.set_tonality(partition.tonality)
    
  def note_string_id(self, note):
    return self._calc_string_id(note.value)
  
  def _calc_string_id(self, value):
    value = abs(value)
    for i, string in enumerate(self.strings): # Search an exact match
      if string.base_note == value: return i
      
    value2 = value % 12
    for i, string in enumerate(self.strings): # Search a string with the right tone
      if (string.base_note % 12) == value2: return i
      
    for i, string in enumerate(self.strings): # Search a string with a #
      if string.base_note == value - 1: return i
      
    for i, string in enumerate(self.strings): # Search a string with the right tone and a #
      if (string.base_note % 12) == value2 - 1: return i
    return len(self.strings) - 1
  
  def get_drawer(self, canvas, compact = False):
    from songwrite3.plugins.lyre.drawer import LyreDrawer
    return LyreDrawer(canvas, self.partition, compact)
  
  def get_icon_filename(self): return "lyre.png"
  
  def get_type(self):
    if   len(self.strings) == 6: return model.TablatureView.get_type(self, SixStringLyreView)
    else:                        return model.TablatureView.get_type(self, SevenStringLyreView)
    
  def set_tonality(self, tonality):
    # Remove previous tonality first
    alterations = model.TONALITIES[self.tonality]
    if alterations:
      if alterations[0] == model.DIESES[0]: alteration_type =  1 # diese
      else:                                 alteration_type = -1 # bemol
    else:                                   alteration_type =  0
    
    for string in self.strings:
      for alteration in alterations:
        if (string.base_note % 12) - alteration_type == alteration:
          string.base_note -= alteration_type
          break
        
    self.tonality = tonality
    
    alterations = model.TONALITIES[tonality]
    if alterations:
      if alterations[0] == model.DIESES[0]: alteration_type =  1 # diese
      else:                                 alteration_type = -1 # bemol
    else:                                   alteration_type =  0
    
    for string in self.strings:
      for alteration in alterations:
        if (string.base_note % 12) == alteration:
          string.base_note += alteration_type
          break
        
  def __xml__(self, xml = None, context = None): model.TablatureView.__xml__(self, xml, context, type = "lyre")


_LYRE_STRING_COLORS = [
  "red"  , #  0
  "red"  , #  1
  "black", #  2
  "black", #  3
  "black", #  4
  "blue" , #  5
  "blue" , #  6
  "black", #  7
  "black", #  8
  "black", #  9
  "black", # 10
  "black", # 11
  ]
class LyreString(model.TablatureString):
  def color(self): return _LYRE_STRING_COLORS[self.base_note % 12]
model.LyreString = LyreString

class SixStringLyreView(LyreView):
  def __init__(self, partition, name = ""):
    super(SixStringLyreView, self).__init__(partition, name or _("Lyre"), self.new_strings())
  @classmethod
  def new_strings(Class):
    return[LyreString(64, -1), LyreString(62, -1), LyreString(60, -1), LyreString(59, -1), LyreString(57, -1), LyreString(55, -1)]
  
class SevenStringLyreView(LyreView):
  def __init__(self, partition, name = ""):
    super(SevenStringLyreView, self).__init__(partition, name or _("Lyre"), self.new_strings())
  @classmethod
  def new_strings(Class):
    return[LyreString(64, -1), LyreString(62, -1), LyreString(60, -1), LyreString(59, -1), LyreString(57, -1), LyreString(55, -1), LyreString(53, -1)]
  

  
plugins.ViewPlugin(LyreView           , LyreString, "lyre"             , None)
plugins.ViewPlugin(SixStringLyreView  , LyreString, "six_string_lyre"  , "tab")
plugins.ViewPlugin(SevenStringLyreView, LyreString, "seven_string_lyre", "tab")



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2001-2011 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Semintone / interval
#  0 base
#  1 nine_flat
#  2 nine
#  3 three_minor
#  4 three_major
#  5 four
#  6 five_flat
#  7 five
#  8 five_sharp
#  9 six
# 10 seven_minor
# 11 seven_major

import sys
import songwrite3
import songwrite3.globdef as globdef
import songwrite3.model   as model

class Chord(object):
  def __init__(self, base, identified_notes, type, tonality = "C"):
    self.base             = base
    self.identified_notes = identified_notes
    self.type             = type
    if len(identified_notes) == 1: # Bass
      self.name           = "%s %s" % (songwrite3.model.note_label(base, 0, tonality, globdef.config.ENGLISH_CHORD_NAME).upper(), type)
    else: # Chord
      self.name           = "%s %s" % (songwrite3.model.note_label(base, 0, tonality, globdef.config.ENGLISH_CHORD_NAME), type)
      
class ChordManager(object):
  def __init__(self, partition, reverse_penality = -65):
    self.partition        = partition
    self.reverse_penality = reverse_penality
    self.caches           = {}
    self.default_thirds   = None
    
  def _identify_relative_notes(self, relative_notes):
    if   0 in relative_notes: first = 0; relative_notes.remove(0)
    else:                     first = None
    
    if   4 in relative_notes: third = 4; relative_notes.remove(4) # Major
    elif 3 in relative_notes: third = 3; relative_notes.remove(3) # Minor
    else:                     third = None
    
    if   7 in relative_notes: fifth = 7; relative_notes.remove(7) # Fifth
    elif 8 in relative_notes: fifth = 8; relative_notes.remove(8) # Sharp fifth / augmented
    elif 6 in relative_notes: fifth = 6; relative_notes.remove(6) # Flat fifth / diminished
    else:                     fifth = None
    
    seventh = None
    if   10 in relative_notes: seventh = 10; relative_notes.remove(10) # Minor seventh
    elif 11 in relative_notes: seventh = 11; relative_notes.remove(11) # Major seventh
    elif  9 in relative_notes:
      if (third == 3) and (fifth == 6): # Only with minor third and diminished fifth
        seventh = 9; relative_notes.remove(9) # Diminished seventh
        
    if   5 in relative_notes: fourth = 5; relative_notes.remove(5) # Sus4 / Eleventh
    else:                     fourth = None
    
    if   2 in relative_notes: second = 2; relative_notes.remove(2) # Sus2 / Nineth
    elif 1 in relative_notes: second = 1; relative_notes.remove(1) # Flat nineth
    else:                     second = None
    
    if   9 in relative_notes: sixth = 9; relative_notes.remove(9) # Sixth
    else:                     sixth = None
    
    return first, second, third, fourth, fifth, sixth, seventh
  
  
  def _name_identified_notes(self, base, first, second, third, fourth, fifth, sixth, seventh):
    if first is None: return 0, ""
    t0 = t2 = t3 = t4 = t5 = t6 = t7 = t9 = t11 = ""
    lik = 0

    # 3th
    if   third == 4: t0 = "M"; t3= ""; lik += 30
    elif third == 3: t0 = "m"; t3= ""; lik += 30
    else:            t3 = "omit3"

    # 5th
    if   fifth == 6: t5 = "5♭" ; lik += 10
    elif fifth == 7:
      lik += 50
      if t3 == "omit3": t5 = "5"; t3 = ""
      else:              t5 = ""
    elif fifth == 8: t5 = "5♯"; lik += 10
    else:            t5 = "omit5"

    if (t0 == "M") and (t5 == "5♯"): t0 = "aug" ; t5 = ""; lik += 15
    if (t0 == "m") and (t5 == "5♭" ): t0 = "dim"; t5 = ""; lik += 10

    # 7th
    if   seventh ==  9: t7 = "7dim"
    elif seventh == 10: t7 = "7m"     ; lik += 20
    elif seventh == 11: t7 = "7M"     ; lik += 10

    if   (t0 == "M"  ) and (t7 == "7M"  ): t0 = "7M"      ; t7 = ""; lik += 10
    elif (t0 == "m"  ) and (t7 == "7m"  ): t0 = "m 7"     ; t7 = ""; lik += 10
    elif (t0 == "M"  ) and (t7 == "7m"  ): t0 = "7"       ; t7 = ""; lik += 10
    elif (t0 == "dim") and (t7 == "7m"  ): t0 = "half dim"; t7 = ""
    elif (t0 == "dim") and (t7 == "7dim"): t0 = "dim 7"   ; t7 = ""
    elif (t0 == "m"  ) and (t7 == "7M"  ): t0 = "m 7M"    ; t7 = ""
    elif (t0 == "aug") and (t7 == "7M"  ): t0 = "aug 7M"  ; t7 = ""
    elif (t0 == "aug") and (t7 == "7m"  ): t0 = "aug 7"   ; t7 = ""
    elif (t0 == ""   ) and (t7 == "7m"  ): t0 = "7"       ; t7 = ""
    elif (t0 == ""   ) and (t7 == "7M"  ): t0 = "7M"      ; t7 = ""

    # 9th and 11th
    if not third:
      if   second == 2: t2 = "sus2" ; t3 = ""; lik += 10
      elif second == 1: t2 = "sus2♭"; t3 = ""; lik -= 20
      if   fourth == 5: t4 = "sus4" ; t3 = ""; lik += 20
    else:
      if   second == 2: t9  = "add9" ; lik += 20
      elif second == 1: t9  = "add9♭"; lik +=  0
      if   fourth == 5: t11 = "add11"; lik += 10

    if (t0 == "7M"    ) and (t9  == "add9" ): t0 = "9M"    ; t9 = ""
    if (t0 == "m 7"   ) and (t9  == "add9" ): t0 = "m 9"   ; t9 = ""
    if (t0 == "7"     ) and (t9  == "add9" ): t0 = "9"     ; t9 = ""
    if (t0 == "aug 7" ) and (t9  == "add9" ): t0 = "aug 9" ; t9 = ""
    if (t0 == "aug 7M") and (t9  == "add9" ): t0 = "aug 9M"; t9 = ""
    if (t0 == "m 7M"  ) and (t9  == "add9" ): t0 = "m 9M"  ; t9 = ""

    if (t0 == "7M"    ) and (t11 == "add11"): t0 = "11M"   ; t11 = ""; lik += 5
    if (t0 == "m 7"   ) and (t11 == "add11"): t0 = "m 11"  ; t11 = ""; lik += 5
    if (t0 == "7"     ) and (t11 == "add11"): t0 = "11"    ; t11 = ""; lik += 5
    if (t0 == "m 7M"  ) and (t11 == "add11"): t0 = "m 11M" ; t11 = ""; lik += 5

    if (t0 == "9M"    ) and (t11 == "add11"): t0 = "11M"   ; t11 = ""; lik += 5
    if (t0 == "m 9"   ) and (t11 == "add11"): t0 = "m 11"  ; t11 = ""; lik += 5
    if (t0 == "9"     ) and (t11 == "add11"): t0 = "11"    ; t11 = ""; lik += 5
    if (t0 == "m 9M"  ) and (t11 == "add11"): t0 = "m 11M" ; t11 = ""; lik += 5

    # 6th
    if sixth == 9: t6 = "add6"; lik -= 20
    
    if (t0 == "m") and (t6 == "add6"): t0 = "m 6"; t6 = ""; lik += 10
    
    if self.default_thirds:
      default = self.default_thirds.get(base % 12)
      if default:
        if   t0 == default:                   t0 = ""
        elif t0.startswith("%s " % default): t0 = t0[len(default) + 1:]
    else:
      if t0 == "M": t0 = ""
      
    name = " ".join([_(x) for x in [t0, t2, t3, t4, t5, t6, t7, t9, t11] if x])

    if name.count(_("omit")) > 1: return 0, ""

    return lik, name


  def identify(self, values):
    if values in self.caches:
      return self.caches[values]
    else:
      chord = self.caches[values] = self._identify(values)
      return chord
    
  def _identify(self, values):
    lower_value = min(values)
    best_lik    = 0
    best_base   = None
    best_name   = ""
    best_notes  = None
    for base in values:
      relative_notes = set([(value - base) % 12 for value in values])
      identified_notes = self._identify_relative_notes(relative_notes)
      lik, name = self._name_identified_notes(base, *identified_notes)
      if name is not None:
        if base != lower_value:
          lik += self.reverse_penality
        if lik > best_lik:
          best_lik   = lik
          best_base  = base
          best_name  = name
          best_notes = identified_notes
    if best_base is None:
      print("XXX pas d'accord pour", values, file = sys.stderr)
      return None
    return Chord(best_base, best_notes, best_name, self.partition.tonality)
  
  def create(self, base, second = 0, third = 4, fourth = 0, fifth = 7, sixth = 0, seventh = 0):
    pass
  
  def choose_third(self, partition, base):
    if ((base + 4 - model.OFFSETS[partition.tonality]) % 12) in model.NOTES: return 4
    else:                                                                    return 3
    
class TablatureChordManager(ChordManager):
  def __init__(self, partition):
    ChordManager.__init__(self, partition)
    
  def create(self, base, second = 0, third = 4, fourth = 0, fifth = 7, sixth = 0, seventh = 0, barred = 0):
    if barred >= 14: return None
    base = base % 12
    allowed_values = set([base, (base + second) % 12, (base + third) % 12, (base + fourth) % 12, (base + fifth) % 12, (base + sixth) % 12, (base + seventh) % 12])
    needed_values  = set(allowed_values)
    values         = []
    frets          = []
    base_placed    = 0
    strings        = list(reversed(self.partition.view.strings))
    max_string_id  = len(strings) - 1
    for i in range(len(strings)):
      if not base_placed:
        value = self._relative_note_on_string(strings[i].base_note, base, barred)
        fret = value - strings[i].base_note
        if fret < strings[i + 1].base_note - strings[i].base_note + barred:
          values.append((value, max_string_id - i))
          frets .append(fret)
          needed_values.discard(base)
          base_placed = 1
      else:
        value = min([self._relative_note_on_string(strings[i].base_note, allowed_value, barred) for allowed_value in allowed_values])
        values.append((value, max_string_id - i))
        needed_values.discard(value % 12)
        frets .append(value - strings[i].base_note)
        
    barred = min(frets)
    if needed_values or self._frets_impossible(frets, barred): return self.create(base, second, third, fourth, fifth, sixth, seventh, barred + 1)
    
    frets = "".join([str(fret) for fret in frets])
    frets = "X" * (len(strings) - len(frets)) + frets
    return values, frets
  
  def _frets_impossible(self, frets, barred = 0):
    if len(frets) - frets.count(barred) > 4: return 1
    if barred and (max(frets) - barred > 2): return 1
    if (not barred) and (max(frets) - barred > 3): return 1
    if (frets.count(barred + 3) >= 2) and (frets.count(barred + 1) >= 2): return 1
    if (frets.count(barred + 4) >= 2) and (frets.count(barred + 2) >= 2): return 1
    if (frets.count(barred + 1) >= 4): return 1
    if (frets.count(barred + 2) >= 4): return 1
    if (frets.count(barred + 3) >= 4): return 1
    if (frets.count(barred + 4) >= 4): return 1
    if len(frets) < 3: return 1
    
  def _relative_note_on_string(self, string_base_note, relative_note, barred = 0):
    while relative_note < string_base_note + barred: relative_note += 12
    return relative_note

class PianoChordManager(ChordManager):
  def __init__(self, partition):
    ChordManager.__init__(self, partition)
    
  def create(self, base, second = 0, third = 4, fourth = 0, fifth = 7, sixth = 0, seventh = 0):
    base   = ((base + 7) % 12) + 41
    values = list(set([(base          , -1),
                       (base + second , -1),
                       (base + third  , -1),
                       (base + fourth , -1),
                       (base + fifth  , -1),
                       (base + sixth  , -1),
                       (base + seventh, -1),
                       (base +      12, -1),
                       ]))
    return values, ""

class AccordionChordManager(ChordManager):
  def __init__(self, partition, instrument_tonality = "G/C"):
    ChordManager.__init__(self, partition)
    self.instrument_tonality = instrument_tonality
    if   instrument_tonality == "G/C": delta = 0
    elif instrument_tonality == "A/D": delta = 2
    self.default_thirds = {
      ( 0 + delta) % 12 : "M", # Do
      ( 2 + delta) % 12 : "M", # Ré
      ( 4 + delta) % 12 : "m", # Mi
      ( 5 + delta) % 12 : "M", # Fa
      ( 7 + delta) % 12 : "M", # Sol
      ( 9 + delta) % 12 : "m", # La
      }
    
  def _identify(self, values):
    if len(values) == 1:
      return Chord(tuple(values)[0], [0], "", self.partition.tonality)
    else:
      chord = ChordManager._identify(self, values)
      words = chord.name.split(None, 1)
      if len(words) == 1: chord.name = chord.name.lower()
      else:               chord.name = "%s %s" % (words[0].lower(), words[1])
      return chord
    
  def create(self, base, second = 0, third = 4, fourth = 0, fifth = 7, sixth = 0, seventh = 0):
    if (second == 0) and (third == 0) and (fourth == 0) and (fifth == 0) and (sixth == 0) and (seventh == 0):
      base   = ((base + 7) % 12) + 41
      values = [(base, -1)]
    else:
      base   = ((base + 7) % 12) + 41
      values = list(set([(base          , -1),
                         (base + second , -1),
                         (base + third  , -1),
                         (base + fourth , -1),
                         (base + fifth  , -1),
                         (base + sixth  , -1),
                         (base + seventh, -1),
                         ]))
    return values, ""
  
  def choose_third(self, partition, base):
    default = self.default_thirds.get(base % 12)
    if   default == "M": return 4
    elif default == "m": return 3
    if ((base + 4 - model.OFFSETS[partition.tonality]) % 12) in model.NOTES: return 4
    else:                                                                    return 3
    return ChordManager.choose_third(self, partition, base)
    

class LyreChordManager(ChordManager):
  def __init__(self, partition):
    ChordManager.__init__(self, partition, reverse_penality = 0)
    
  def create(self, base, second = 0, third = 4, fourth = 0, fifth = 7, sixth = 0, seventh = 0):
    base           = base % 12
    allowed_values = set([base, (base + second) % 12, (base + third) % 12, (base + fourth) % 12, (base + fifth) % 12, (base + sixth) % 12, (base + seventh) % 12])
    values         = []
    frets          = ""
    strings        = list(reversed(self.partition.view.strings))
    max_string_id  = len(strings) - 1
    for i in range(len(strings)):
      if strings[i].base_note % 12 in allowed_values:
        values.append((strings[i].base_note, max_string_id - i))
        frets += "0"
      else:
        frets += "X"
    return values, frets
    


if __name__ == "__main__":
  import songwrite3, songwrite3.model, songwrite3.plugins.lyre
  partition = songwrite3.model.Partition(None)
  partition.set_view_type(songwrite3.model.GuitarView)
  manager = TablatureChordManager(partition)
  #partition.set_view_type(songwrite3.plugins.lyre.SevenStringLyreView)
  #manager = LyreChordManager(partition)
  #notes = frozenset([
  #  36,
  #  36 + 2,
  #  36 + 4,
  #  36 + 7,
  #  ])
  notes = frozenset([64, 52, 60, 48, 55])
  chord = manager.identify(notes)
  print(chord.name)
  print()
  #for i in range(12):
  #  print songwrite3.model.note_label(i) + "  \t" + manager.create(i)[1], manager.create(i, third = 3)[1], manager.create(i, seventh = 10)[1], manager.create(i, third = 3, seventh = 10)[1]
  
  



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2007-2015 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from io import StringIO, BytesIO

import songwrite3.plugins as plugins
import songwrite3.midi    as midi


class MidiExportPlugin(plugins.ExportPlugin):
  def __init__(self):
    plugins.ExportPlugin.__init__(self, "MIDI", [".mid", ".midi"])
    
  def export_to_string(self, song):
    return midi.song_2_midi(song)
    
MidiExportPlugin()


class MidiImportPlugin(plugins.ImportPlugin):
  def __init__(self):
    plugins.ImportPlugin.__init__(self, "MIDI", [".mid", ".midi"], binary = True)
    
  def import_from_string(self, data):
    import songwrite3.plugins.midi.importer as importer
    return importer.parse(BytesIO(data))
    
MidiImportPlugin()


class RichMidiExportPlugin(plugins.ExportPlugin):
  def __init__(self):
    plugins.ExportPlugin.__init__(self, "RichMIDI", [".mid", ".midi"])
    
  def export_to_string(self, song):
    return midi.song_2_midi(song, rich_midi_tablature = 1)
    
RichMidiExportPlugin()


class RichMidiImportPlugin(plugins.ImportPlugin):
  def __init__(self):
    plugins.ImportPlugin.__init__(self, "RichMIDI", [".mid", ".midi"], binary = True)
    
  def import_from_string(self, data):
    import songwrite3.plugins.midi.importer as importer
    return importer.parse(BytesIO(data), rich_midi_tablature = 1)
    
RichMidiImportPlugin()



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2012 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtGui     as qtgui

import songwrite3.canvas as canvas



class LuteDrawer(canvas.TablatureDrawer):
  def __init__(self, canvas_, partition, compact = False):
    canvas.TablatureDrawer.__init__(self, canvas_, partition, compact)
    
  def on_key_press(self, event):
    keyval = event.key()
    if   qtcore.Qt.Key_A <= keyval <= qtcore.Qt.Key_N: # a-n
      self.canvas.on_number_typed(str(keyval - qtcore.Qt.Key_A))
      return 1




# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2007-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, atexit

import songwrite3.plugins as plugins
import songwrite3.model   as model
import songwrite3.globdef as globdef


PREVIEW_TMP_FILE = ""

def remove_preview_tmp_file():
  global PREVIEW_TMP_FILE
  if PREVIEW_TMP_FILE:
    os.unlink(PREVIEW_TMP_FILE)
    PREVIEW_TMP_FILE = ""
atexit.register(remove_preview_tmp_file)


class ExportPluginPDF(plugins.ExportPlugin):
  def __init__(self):
    plugins.ExportPlugin.__init__(self, "PDF", [".pdf"], 1, 1)
    
  def export_to_string(self, song):
    import songwrite3.plugins.pdf.pdf_latex as pdf_latex
    return pdf_latex.pdfy(song)
  
  def print_preview(self, song):
    import songwrite3.plugins.pdf.pdf_latex as pdf_latex
    global PREVIEW_TMP_FILE
    
    if PREVIEW_TMP_FILE: remove_preview_tmp_file()
    
    pdf = pdf_latex.pdfy(song)
    
    import subprocess

    pdf_command = globdef.config.get_preview_command_pdf()
    if "%s" in pdf_command:
      import tempfile
      fid, PREVIEW_TMP_FILE = tempfile.mkstemp(suffix = ".pdf", text = 0)
      with open(PREVIEW_TMP_FILE, "wb") as f: f.write(pdf)
      command = pdf_command % PREVIEW_TMP_FILE
      print("Running '%s'" % command, file = sys.stderr)
      p = subprocess.Popen(command, shell = True, close_fds = True)
    else:
      print("Running '%s'" % pdf_command, file = sys.stderr)
      p = subprocess.Popen(pdf_command, shell = True, stdin = subprocess.PIPE, close_fds = True)
      p.stdin.write(pdf)
      p.stdin.close()
      
ExportPluginPDF()



# -*- coding: utf-8 -*-

# SongWrite
# Copyright (C) 2001-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This is the parser/reader for StemML.

import sys, xml.sax as sax, xml.sax.handler as handler
from io import StringIO
import codecs

import songwrite3.model   as model
import songwrite3.plugins as plugins

def parse(file):
  parser = sax.make_parser()
  song_handler = SongHandler(parser, file)
  parser.setContentHandler(song_handler)
  parser.parse(file)
  
  if   hasattr(song_handler, "songbook"): return song_handler.songbook
  elif hasattr(song_handler, "notes"   ): return song_handler.click_x, song_handler.notes
  else:                                   return song_handler.song
  
  
class SongHandler(handler.ContentHandler):
  def __init__(self, parser, file):
    self.parser = parser
    self.current    = ""
    self.need_strip = 0
    self.file = file
    
  def startElement(self, name, attrs):
    if   name == "bar":
      self.bar_attrs.update(attrs)
      rythm = self.bar_attrs["rythm"]
      i = rythm.index("/")
      if self.song.mesures:
        self.song.mesures.append(model.Mesure(self.song, self.song.mesures[-1].end_time(), int(self.bar_attrs["tempo"]), int(rythm[:i]), int(rythm[i + 1:]), int(self.bar_attrs["syncope"])))
      else:
        self.song.mesures.append(model.Mesure(self.song, 0,                                int(self.bar_attrs["tempo"]), int(rythm[:i]), int(rythm[i + 1:]), int(self.bar_attrs["syncope"])))
        
    elif name == "play":
      self.song.playlist.playlist_items.append(model.PlaylistItem(self.song.playlist, int(attrs["from"]), int(attrs["to"])))
      
    elif name == "partition":
      partition_handler = PartitionHandler(self.parser, self, self.song, attrs)
      self.parser.setContentHandler(partition_handler)
      
    elif name == "lyrics":
      lyrics_handler = LyricsHandler(self.parser, self, self.song, attrs)
      self.parser.setContentHandler(lyrics_handler)
      
    elif name == "bars": pass
    elif name == "playlist": pass
    elif name == "song":
      self.bar_attrs = {"tempo" : 60, "syncope" : 0 , "rythm" : "4/4"}
      self.song = self.obj = model.Song()
      self.song.mesures      *= 0
      self.song.version       = attrs.get("version", model.VERSION)
      self.song.lang          = attrs.get("lang", "en")[:2]
      self.song.print_nb_mesures_per_line = int(attrs.get("print_nb_mesures_per_line", 4))
      self.song.printfontsize = int(attrs.get("printfontsize", 16))
      self.song.print_lyrics_columns = int(attrs.get("print_lyrics_columns", 1))
      if attrs.get("print_lyrics_on_two_columns", 0):
        self.song.print_lyrics_columns = 2
      self.song.set_filename(self.file)
    elif name == "songfile":
      self.songbook.add_song_ref(model.SongRef(self.songbook, attrs.get("filename"), attrs.get("title")))
    elif name == "songbook":
      self.songbook = self.obj = model.Songbook(self.file, attrs.get("title", ""), attrs.get("authors", ""), attrs.get("comments", ""))
      self.songbook.version = attrs.get("version", model.VERSION)
    elif name == "notes":
      partition_handler = PartitionHandler(self.parser, self, None, attrs)
      self.parser.setContentHandler(partition_handler)
      self.click_x = partition_handler.partition.click_x
      self.notes   = partition_handler.partition.notes
    else: self.current = name
    
  def endElement(self, name):
    if   name == "song":
      self.song.playlist.analyse()
      
    if self.need_strip:
      setattr(self.obj, self.current, getattr(self.obj, self.current).strip())
      self.need_strip = 0
    self.current = ""
    
  def characters(self, content):
    if   self.current == "": pass
    else:
      setattr(self.obj, self.current, getattr(self.obj, self.current, "") + content)
      self.need_strip = 1 # Will strip when the tag ends.


class PartitionHandler(handler.ContentHandler):
  def __init__(self, parser, previous_handler, song, attrs):
    self.parser               = parser
    self.previous_handler     = previous_handler
    self.partition            = model.Partition(song)
    #if "g8" in attrs: self.partition.g8 = attrs["g8"]
    for attr, value in attrs.items():
      try: value = int(value)
      except:
        try: value = float(value)
        except:
          if   value == "True" : value = 1
          elif value == "False": value = 0
      setattr(self.partition, attr, value)

    if song: song.partitions.append(self.partition)
    self.current     = ""
    self.id_to_notes = {}
    self.note_setter = {
      "time"        : None,
      "duration"    : None,
      "pitch"       : None,
      "volume"      : None,
      "linked_to"   : self.note_set_linked_to,
      "fx"          : self.note_set_fx,
      "link_fx"     : self.note_set_link_fx,
      "duration_fx" : self.note_set_duration_fx,
      "strum_dir_fx": self.note_set_strum_dir_fx,
      "appoggiatura": self.note_set_appoggiatura,
      "id"          : self.note_set_id,
      "string"      : self.note_set_string,
      "bend_pitch"  : self.note_set_bend_pitch,
      "label"       : self.note_set_label,
      }
    self.notes_linked_to = []
    self.view_hidden     = 0
    self.need_strip      = 0
    
  def note_set_id(self, note, id): self.id_to_notes[id] = note
  
  def note_set_linked_to(self, note, id):
    # Cannot be done yet, since the following notes are not parsed yet !
    # We put this request somewhere and treat it later.
    self.notes_linked_to.append((note, id))
    
  def note_set_fx(self, note, fx):
    if (fx == "link") or (fx == "slide") or (fx == "hammer") or (fx == "pull"):
      self.note_set_link_fx(note, fx)
    else:
      note.fx = fx
      
  def note_set_link_fx(self, note, fx):
    if (fx == "hammer") or (fx == "pull"): fx = "link"
    note.link_fx = fx
    
  def note_set_duration_fx(self, note, fx):
    note.duration_fx = fx
    
  def note_set_appoggiatura(self, note, appoggiatura):
    if appoggiatura: note.duration_fx = "appoggiatura"
    
  def note_set_string(self, note, string_id):
    note.string_id = int(string_id)
    
  def note_set_label(self, note, label):
    note.label = label
    
  def note_set_bend_pitch(self, note, bend_pitch):
    note.bend_pitch = float(bend_pitch)
    
  def note_set_strum_dir_fx(self, note, fx):
    note.strum_dir_fx = fx
    
  def startElement(self, name, attrs):
    if   name == "note":
      note = model.Note(self.partition,
                        int(round(float(attrs["time"    ]) * 96.0)),
                        int(round(float(attrs["duration"]) * 96.0)),
                        int(attrs["pitch"]),
                        int(attrs.get("volume") or self.partition.notes[-1].volume),
                        )
      for attr in attrs.keys():
        setter = self.note_setter.get(attr, 1)
        if   setter is 1:
          value = attrs[attr]
          try: value = int(value)
          except:
            try: value = float(value)
            except:
              if   value == "True" : value = 1
              elif value == "False": value = 0
          setattr(note, attr, value)
        elif setter: setter(note, attrs[attr])
        
      self.partition.notes.append(note)
      
    elif name == "notes"  : pass
    elif name == "strings": pass
    elif name == "string" :
      if attrs.get("type"):
        self.partition.view.strings.append(getattr(model, attrs.get("type"))(int(attrs.get("pitch") or attrs.get("patch"))))
      else:
        self.partition.view.strings.append(self._string_class(int(attrs.get("pitch") or attrs.get("patch"))))
        
    elif name == "view":
      if   attrs["type"] == "tablature":
        self.partition.view = model.TablatureView(self.partition, "", [])
        self._string_class = model.TablatureString
        
      elif attrs["type"] == "drums":
        self.partition.view = model.DrumsView(self.partition, "", [])
        self._string_class = model.DrumString
        
      elif attrs["type"] == "staff":
        self.partition.view = model.StaffView(self.partition, "")
        self._string_class = None
        
      else:
        self.partition.view, self._string_class = plugins.load_view(self.partition, attrs)
        
      if attrs.get("hidden"): self.partition.view.visible = 0
      
    else: self.current = name
    
  def endElement(self, name):
    if   (name == "partition") or (name == "notes"):
      for note, id in self.notes_linked_to:
        other = self.id_to_notes.get(id, None)
        note .linked_to = other
        if other: other.linked_from = note
        
      self.parser.setContentHandler(self.previous_handler)
        
    elif name == self.current:
      if self.need_strip:
        setattr(self.partition, self.current, getattr(self.partition, self.current).strip())
        self.need_strip = 0
      self.current = ""
      
  def characters(self, content):
    if   self.current == ""    : pass
    else:
      setattr(self.partition, self.current, getattr(self.partition, self.current, "") + content)
      self.need_strip = 1 # Will strip when the tag ends.
      
class LyricsHandler(handler.ContentHandler):
  def __init__(self, parser, previous_handler, song, attrs):
    self.parser           = parser
    self.previous_handler = previous_handler
    self.lyrics           = model.Lyrics(song)
    self.lyrics.show_all_lines_on_melody   = int(attrs.get("show_all_lines_on_melody"  , 0))
    self.lyrics.show_all_lines_after_score = int(attrs.get("show_all_lines_after_score", 1))
    song.partitions.append(self.lyrics)
    self.texts      = []
    self.current    = ""
    self.need_strip = 0
    
  def startElement(self, name, attrs):
    if   name == "br":       self.texts.append("\\\\")
    elif name == "br-verse": self.texts.append("\n")
    else: self.current = name
    
  def endElement(self, name):
    if   name == "lyrics":
      self.lyrics.text = "".join(self.texts).rstrip()
      self.parser.setContentHandler(self.previous_handler)
      
    elif name == self.current:
      if self.need_strip:
        setattr(self.lyrics, self.current, getattr(self.lyrics, self.current).strip())
        self.need_strip = 0
      self.current = ""
      
  def characters(self, content):
    if   self.current == ""    : pass
    elif self.current == "text":
      self.texts.append(content.replace("\n", ""))
    else:
      setattr(self.lyrics, self.current, getattr(self.lyrics, self.current, "") + content)
      self.need_strip = 1 # Will strip when the tag ends.





# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2007 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import songwrite3.model   as model
import songwrite3.plugins as plugins


class ExportPlugin(plugins.ExportPlugin):
  def __init__(self):
    plugins.ExportPlugin.__init__(self, "TextTab", [".txt"], 1, 1)
    
  def export_to_string(self, song):
    import songwrite3.plugins.texttab.texttab as texttab
    if isinstance(song, model.Songbook):
      return "\n\n\n\n".join(texttab.texttab(song_ref.get_song()) for song_ref in song.song_refs)
    else:
      return texttab.texttab(song)
    
ExportPlugin()

class ImportPlugin(plugins.ImportPlugin):
  def __init__(self):
    plugins.ImportPlugin.__init__(self, "TextTab", [".txt"])
    
  def import_from(self, filename):
    import songwrite3.plugins.texttab.texttab as texttab
    return texttab.parse(open(filename).read())
  
ImportPlugin()



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2007-2011 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import songwrite3.model   as model
import songwrite3.plugins as plugins

plugins.Plugin("additional_instruments")

class Banjo5GTablatureString(model.TablatureString):
  def value_2_text(self, note):
    if note.value == self.base_note: return "0"
    return unicode(note.value - self.base_note - note.partition.capo + 5)
  
  def text_2_value(self, note, text):
    text = int(text)
    if text <= 5: return self.base_note + note.partition.capo
    return text + self.base_note + note.partition.capo - 5
  
model.Banjo5GTablatureString = Banjo5GTablatureString

class Banjo5GView(model.TablatureView):
  default_instrument = 105
  def __init__(self, partition, name = ""):
    super(Banjo5GView, self).__init__(partition, name or _("Banjo 5G"), self.new_strings())
  @classmethod
  def new_strings(Class):
    return [model.TablatureString(62, -1), model.TablatureString(59, -1), model.TablatureString(55, -1), model.TablatureString(50, -1), Banjo5GTablatureString(67, -1)]
model.VIEWS["tab"].append(Banjo5GView)

class UkuleleView(model.TablatureView):
  default_instrument = 27
  def __init__(self, partition, name = ""):
    super(UkuleleView, self).__init__(partition, name or _("Ukulele"), self.new_strings())
  @classmethod
  def new_strings(Class):
    return [model.TablatureString(69, -1), model.TablatureString(64, -1), model.TablatureString(60, -1), model.TablatureString(55, -1)]
model.VIEWS["tab"].append(UkuleleView)

class MandolinView(model.TablatureView):
  default_instrument = 105
  def __init__(self, partition, name = ""):
    super(MandolinView, self).__init__(partition, name or _("Mandolin"), self.new_strings())
  @classmethod
  def new_strings(Class):
    return [model.TablatureString(76, -1), model.TablatureString(69, -1), model.TablatureString(62, -1), model.TablatureString(55, -1)]
model.VIEWS["tab"].append(MandolinView)

class KoyabuView(model.TablatureView):
  default_instrument = 27
  def __init__(self, partition, name = ""):
    super(KoyabuView, self).__init__(partition, name or _("Koyabu board"), self.new_strings())
  @classmethod
  def new_strings(Class):
    return [
    model.TablatureString(24, -1),
    model.TablatureString(31, -1),
    model.TablatureString(38, -1),
    model.TablatureString(45, -1),
    model.TablatureString(52, -1),
    model.TablatureString(59, -1),
    model.TablatureString(64, -1),
    model.TablatureString(57, -1),
    model.TablatureString(52, -1),
    model.TablatureString(47, -1),
    model.TablatureString(42, -1),
    model.TablatureString(37, -1),
    ]
model.VIEWS["tab"].append(KoyabuView)


class LuteView(model.TablatureView):
  default_instrument = 27
  def __init__(self, partition, name = ""):
    super(LuteView, self).__init__(partition, name or _("Lute"), self.new_strings())
  @classmethod
  def new_strings(Class):
    return [model.LuteTablatureString(69, -1),
            model.LuteTablatureString(64, -1),
            model.LuteTablatureString(60, -1),
            model.LuteTablatureString(55, -1)]

  def get_drawer(self, canvas, compact = False):
    from songwrite3.plugins.additional_instruments.drawer import LuteDrawer
    return LuteDrawer(canvas, self.partition, compact)

model.VIEWS["tab"].append(LuteView)

class LuteTablatureString(model.TablatureString):
  def value_2_text(self, note):
    if note.value == self.base_note: return "a"
    return chr(note.value - self.base_note - note.partition.capo + 97)
  
model.LuteTablatureString = LuteTablatureString



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2001-2015 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, struct
import songwrite3.model as model
import songwrite3.midi  as midi

def parse(file, rich_midi_tablature = 0):
  song = model.Song()
  
  partitions = {}
  def channel2partition(channel):
    return partitions.get(channel) or new_partition(channel)
  def new_partition(channel):
    partition = model.Partition(song)
    song.partitions.append(partition)
    if channel == 9: # Drums
      partition.instrument = 128
      
    partitions[channel] = partition
    return partition
  
  name, length = struct.unpack(">4si", file.read(8))
  if name != b"MThd": raise ValueError("Not a midi file !", file)
  format, nb_tracks, tempo = struct.unpack(">hhh", file.read(6))

  for i in range(nb_tracks):
    name, length = struct.unpack(">4si", file.read(8))
    if name != b"MTrk": raise ValueError("Not a track !", file)
    
    partitions   = {}
    time         = 0
    opened_notes = {}
    string_id    = None
    fx           = ""
    end_at       = file.tell() + length
    while file.tell() < end_at:
      time  = time + midi.readVarLength(file)
      event = struct.unpack(">B", file.read(1))[0]
      
      if   0x80 <= event < 0x90: # Note off
        channel = event - 0x80
        
        value = struct.unpack(">B", file.read(1))[0]
        try:
          note = opened_notes[channel, value]
          note.duration = time - note.time
          del opened_notes[channel, value]
        except:
          print("Warning ! Note off without note on at time %s !" % time, file = sys.stderr)
        file.read(1)
        
      elif 0x90 <= event < 0xA0: # Note on
        channel = event - 0x90
        
        value, volume = struct.unpack(">BB", file.read(2))
        
        if volume == 0: # A Note on with wolume == 0 is a Note off ???
          try:
            note = opened_notes[channel, value]
            note.duration = time - note.time
            del opened_notes[channel, value]
          except:
            print("Warning ! Note off without note on at time %s !" % time, file = sys.stderr)
            
        else:          
          note = opened_notes[channel, value] = model.Note(channel2partition(channel), time, 256, value, volume) # Duration is unknown
          
          if string_id is not None:
            note.string_id = string_id
            string_id      = None
            
          if fx:
            if fx in ("link", "slide"): note.link_fx = fx
            else:                       note.fx      = fx
            fx = ""
            
          channel2partition(channel).add_note(note)
          
      elif 0xA0 <= event < 0xB0:
        print("Warning ! aftertouch not supported !", file = sys.stderr)
        file.read(2)
        
      elif 0xB0 <= event < 0xC0:
        partition = channel2partition(event - 0xB0)
        
        event = struct.unpack(">B", file.read(1))[0]
        if   event == 0x5B: partition.reverb = struct.unpack(">B", file.read(1))[0]
        elif event == 0x5D: partition.chorus = struct.unpack(">B", file.read(1))[0]
        elif event == 0x07: partition.volume = struct.unpack(">B", file.read(1))[0] * 2
          
        else:
          print("Warning ! unknown midi controller : %s, value : %s" % (hex(event), struct.unpack(">B", file.read(1))[0]), file = sys.stderr)
          
      elif 0xC0 <= event < 0xD0:
        partition = channel2partition(event - 0xC0)
        if partition.instrument != 128: # Else it is drums ; an instrument event has no sens for the drums channel but it occurs in some mids...
          partition.instrument = struct.unpack(">B", file.read(1))[0]
          
      elif 0xD0 <= event < 0xE0:
        print("Warning ! aftertouch not supported !", file = sys.stderr)
        file.read(1)
        
      elif 0xE0 <= event < 0xF0:
        print("Warning ! pitchwheel not supported ! Use rich midi if you want to import hammer/bend/....", file = sys.stderr)
        file.read(1)
        file.read(1)
        
      elif event == 0xF0: # System exclusive
        while struct.unpack(">B", file.read(1))[0] != 0xF7: pass
        
      elif event == 0xFF:
        event   = struct.unpack(">B", file.read(1))[0]
        length  = struct.unpack(">B", file.read(1))[0]
        content = file.read(length)
        
        if   event == 0x01: # Comment ?
          if song.comments: song.comments = song.comments + "\n" + content
          else:             song.comments = content
          
        elif event == 0x02: # Copyright ?
          if song.copyright: song.copyright = song.copyright + content
          else:              song.copyright = content
          
        elif event == 0x10: # Rich tab midi event
          channel_id     = struct.unpack(">B", content[0:1])[0]
          partition      = channel2partition(channel_id - 1)
          partition.capo = struct.unpack(">B", content[1:2])[0]
          
          strings   = []
          for c in content[2:]:
            strings.append(model.TablatureString(struct.unpack(">B", c)[0]))
          partition.view = model.TablatureView(partition, "", strings)
          
        elif event == 0x11: # Rich tab midi event
          if rich_midi_tablature:
            string_id = struct.unpack(">B", content[0:1])[0]
            if len(content) > 1:
              spc = struct.unpack(">B", content[1:2])[0]

              if   spc == 0x01: fx = "link"
              elif spc == 0x02: fx = "link"
              elif spc == 0x03: fx = "slide"
              elif spc == 0x04: fx = "slide"
              elif spc == 0x06: fx = "roll"
              elif spc == 0x0A: fx = "tremolo"
              elif spc == 0x0E: fx = "dead"
              elif spc == 0x0C:
                fx = "bend"
                if len(content) > 2: bend_pitch = struct.unpack(">B", content[2:3])[0] * 0.25
                else:                bend_pitch = 0.5
              else: print("Warning ! unknown rich midi special effect :", hex(spc), file = sys.stderr)
            
        elif event == 0x2F: # End of track
          file.seek(end_at)
          break
        
        else:
          print("Warning ! unknow sequence 0xFF", hex(event), "  content (%s bytes) : %s" % (len(content), content), file = sys.stderr)

      else:
        print("Warning ! unknown midi event :", hex(event), file = sys.stderr)
        continue

  for partition in song.partitions:
    if partition.instrument == 128: # Drums
      strings = {}
      for note in partition.notes:
        if not note.value in strings:
          strings[note.value] = model.DrumString(note.value)
          
  for mesure in song.mesures: mesure.tempo = tempo
  
  # Start the song at time == 0
  if song.partitions:
    start = min([min([sys.maxsize] + [note.time for note in partition.notes]) for partition in song.partitions])
    for partition in song.partitions:
      for note in partition.notes:
        note.time = note.time - start
        
      if not partition.view:
        partition.view = model.GuitarView(partition)
        
  return song



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2001-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# More info on Rich Tablature Midi format here:
# http://www.tabledit.com/midi/rich_midi_tablature_format.html

# MIDI info:
# http://wearcam.org/ece516/MIDIspec.htm

import sys, struct, math
import songwrite3.model as model

def apply_syncope(mesure, time):
  if (mesure.rythm1 == 6) and (mesure.rythm2 == 8): # Jig
    t = (time - mesure.time) % 144
    if   t < 48: t = t * 1.2      # First  half, longer
    elif t < 96: t = t * 0.8 + 19.2 # Second half, shorter
                                          # Third  half is normal
    return mesure.time + ((time - mesure.time) // 144) * 144 + int(t)
  
  else: # Ternary syncope
    t = (time - mesure.time) % 96
    if t < 48: t = t * 1.333333333      # First  half, longer
    else:      t = t * 0.666666666 + 32 # Second half, shorter
    return mesure.time + ((time - mesure.time) // 96) * 96 + int(t)
  

class TooManyChannelError(model.SongwriteError): pass

class MidiBaseChannel(object): # A midi channel, for internal use
  def __init__(self, midi, id):
    if id >= 16: raise TooManyChannelError
    self.midi       = midi
    self.special_id = self.id = id
    
  def set_init_events(self, init_events):
    self.init_events = init_events
    self.midi.events.extend(init_events)
    
  def __int__(self): return self.id
  
class MidiChannel(MidiBaseChannel): # A normal midi channel, for internal use
  def __init__(self, midi):
    MidiBaseChannel.__init__(self, midi, midi.next_channel_id)
    midi.next_channel_id += 1
    if midi.next_channel_id == 9: # Channel 9 is reserved for Drums !
      midi.next_channel_id += 1
      
  def fork_for_special_notes(self):
    self.special_id = self.midi.next_channel_id
    if self.special_id >= 16: raise TooManyChannelError
    self.midi.next_channel_id += 1
    
    self.midi.events.extend( # Transpose the initing events so it apply to the new channel -- the special channel must be initialized as the normal was !
      (data[0], data[1], bytes([data[2][0] - self.id + self.special_id]) + data[2][1:])
      for data in self.init_events
      )
    
class Midi(object): # A midifier, for internal use
  def __init__(self, rich_midi_tablature = 0):
    self.next_channel_id     = 0
    self.events              = []
    self.drums_channel       = MidiBaseChannel(self, 9) # Play drums on channel 9 (10 if we start counting at 1)
    self.rich_midi_tablature = rich_midi_tablature
    self.partition_2_channel = {}
    
  def data(self, mesures, playlist = None, trackable = 0):
    self.events.sort()
    
    lasttime = 0
    events2  = []
    
    time_factor = 1.0
    old_tempo   = mesures[0].tempo
    
    real_time = 0
    
    if playlist and playlist.playlist_items:
      events1 = []
      mesure2events = { mesure : [] for mesure in mesures }
      
      for time_event in self.events:
        # Before event is a Midi meta-event that be stay JUST BEFORE event (this is for the Rich Tablature Midi Format)
        if len(time_event) == 3:
          time, mesure, event  = time_event
          source = None
        else: time, mesure, event, source = time_event
        
        if mesure is None:
          events2.append(varLength(0) + event)
          continue
          
        if time < 0: time = 0
        
        if mesure.syncope: time = apply_syncope(mesure, time)
        
        mesure2events[mesure].append((time, event, source))
        lasttime = time
        
      mesuretime = 0
      lasttime   = 0
      for item in playlist.playlist_items:
        for mesure in mesures[item.from_mesure : item.to_mesure + 1]:
          if mesure.tempo != old_tempo:
            time_factor = time_factor * old_tempo / mesure.tempo
            old_tempo = mesure.tempo
            
          for time_event in mesure2events[mesure]:
            time2, event, source = time_event
            time = time2 - mesure.time + mesuretime
            
            delta_time = (time - lasttime) * time_factor
            real_time += delta_time
            
            if self.rich_midi_tablature and source: # Add the rich midi tablature note event JUST BEFORE the event
              if isinstance(source, model.Partition): rich_midi_event = partition_2_rich_midi_events(source, self)
              else:                                   rich_midi_event = note_2_rich_midi_events(source)
              if rich_midi_event: events1.append((real_time, rich_midi_event))
              
            events1.append((real_time, event))
            
            if trackable and isinstance(source, model.Note):
              if 144 <= event[0] < 160:
                label = bytes(str(int(source.time)), "ascii")
                events1.append((real_time, struct.pack(">BB", 0xFF, 0x05) + varLength(1+len(label)) + label + b"\n"))
                
            lasttime = time
            
          mesuretime += mesure.duration
          
      events1.sort()
      lasttime = time
      for time, event in events1:
        events2.append(varLength(int(time - lasttime)) + event)
        lasttime = time
        
    else:
      for time_event in self.events:
        # Before event is a Midi meta-event that be stay JUST BEFORE event (this is for the Rich Tablature Midi Format)
        if len(time_event) == 3:
          time, mesure, event  = time_event
          source = None
        else: time, mesure, event, source = time_event
        
        if time < 0: time = 0
        
        if mesure:
          if mesure.syncope: time = apply_syncope(mesure, time)
            
          if mesure.tempo != old_tempo:
            time_factor = time_factor * old_tempo / mesure.tempo
            old_tempo = mesure.tempo
            
        delta_time = (time - lasttime) * time_factor
        real_time += delta_time
        
        if self.rich_midi_tablature and source: # Add the rich midi tablature note event JUST BEFORE the event
          if isinstance(source, model.Partition): rich_midi_event = partition_2_rich_midi_events(source, self)
          else:                                   rich_midi_event = note_2_rich_midi_events(source)
          
          if rich_midi_event:
            events2.append(varLength(int(delta_time)) + rich_midi_event)
            lasttime = time
            delta_time = 0
        events2.append(varLength(int(delta_time)) + event)
        
        if trackable and isinstance(source, model.Note):
          label = bytes(str(int(source.time)), "ascii")
          events2.append(varLength(0) + struct.pack(">BB", 0xFF, 0x05) + varLength(1+len(label)) + label + b"\n")
          
        lasttime = time
        
    if trackable:
      events2.append(varLength(0) + struct.pack(">BB", 0xFF, 0x05) + varLength(3) + b"-1\n")
    events2.append(varLength(0) + END_TRACK)
    
    return chunk(b"MThd", struct.pack(">hhh", 0, 1, mesures[0].tempo)) + chunk(b"MTrk", b"".join(events2))
    
    

def song_2_midi(self, start_time = 0, end_time = sys.maxsize, rich_midi_tablature = 0, trackable = 0):
  midi = Midi(rich_midi_tablature)
  
  for partition in self.partitions:
    if isinstance(partition, model.Partition):
      #delta = getattr(partition, "capo", 0)
      delta = 0
      for note in partition.notes:
        note._real_value = note.value + delta
  for partition in self.partitions:
    if isinstance(partition, model.Partition):
      partition_2_midi(partition, midi, start_time, end_time)
  if start_time != 0:
    return midi.data(self.mesures, (), trackable)
  else:
    return midi.data(self.mesures, self.playlist, trackable)
    ME_SYSEX_GS_LSB
def partition_2_midi(self, midi, start_time, end_time):
  if not self.muted:
    if self.instrument == 128: # 128 means Drums
      channel = midi.partition_2_channel[self] = midi.drums_channel
      channel.set_init_events((
        (-1, None, struct.pack(">BBB", 0xB0 + channel.id, 0x5B, self.reverb     )), # Reverb
        (-1, None, struct.pack(">BBB", 0xB0 + channel.id, 0x5D, self.chorus     )), # Chorus
        (-1, None, struct.pack(">BBB", 0xB0 + channel.id, 0x07, self.volume // 2)), # Volume
        ))
    else:
      channel = midi.partition_2_channel[self] = MidiChannel(midi)
      channel.set_init_events((
        (-1, None, struct.pack(">BB" , 0xC0 + channel.id, self.instrument), self ), # Instrument selection
        (-1, None, struct.pack(">BBB", 0xB0 + channel.id, 0x5B, self.reverb     )), # Reverb
        (-1, None, struct.pack(">BBB", 0xB0 + channel.id, 0x5D, self.chorus     )), # Chorus
        (-1, None, struct.pack(">BBB", 0xB0 + channel.id, 0x07, self.volume // 2)), # Volume
        ))
      
    # Check if some "special notes" (hammer, ...) are played at the same time that normal notes, and if so, play them on a different channel.
    fx_notes = [note for note in self.notes if note.fx or note.link_fx]
    if fx_notes and not midi.rich_midi_tablature:
      # We must ALWAYS fork since we don't know when a note really ends...
      channel.fork_for_special_notes()
      
    mesure    = self.song.mesures[0]
    mesure_id = 0
    
    if midi.rich_midi_tablature: note_2_events = note_2_midi_events_without_special_effect
    else:                        note_2_events = note_2_midi_events
    
    for note in self.notes:
      while note.time >= mesure.end_time():
        mesure_id += 1
        mesure = self.song.mesures[mesure_id]
        
      midi.events.extend(note_2_events(note, channel, mesure, start_time, end_time))
      
def partition_2_rich_midi_events(self, midi):
  if not hasattr(self.view, "strings"): return None
  channel = midi.partition_2_channel[self]
  return (b"\xff\x10" +
          struct.pack(">BBB", len(self.view.strings) + 2, channel.id + 1, getattr(self, "capo", 0)) +
          b"".join(struct.pack(">B", string.base_note) for string in self.view.strings))



def note_2_midi_events_without_special_effect(self, channel, mesure, start_time, end_time):
  if start_time <= self.time <= end_time:
    return (
      (self.time - start_time                , mesure, struct.pack(">BBB", 0x90 + channel.id, self._real_value, self.volume), self),
      (self.time - start_time + self.duration, mesure, struct.pack(">BBB", 0x80 + channel.id, self._real_value, self.volume)),
      )
  return ()

def note_2_midi_events(self, channel, mesure, start_time, end_time):
  if start_time <= self.time <= end_time:
    if   self.link_fx:
      func = globals().get("_note_2_midi_events_%s" % self.link_fx)
      if func:
        r = func(self, channel, mesure, start_time, end_time)
        if not r is None: return r
    elif self.fx:
      func = globals().get("_note_2_midi_events_%s" % self.fx)
      if func:
        r = func(self, channel, mesure, start_time, end_time)
        if not r is None: return r
    if self.linked_from: return ()
    real_time = self.real_time()
    if getattr(self.partition, "let_ring", 0):
      next = self.next()
      if next: duration = min(next.time - self.time, 384)
      else:    duration = 384
    else: duration = self.duration
    return (
      (real_time - start_time           , mesure, struct.pack(">BBB", 0x90 + channel.id, self._real_value, self.volume), self),
      (real_time - start_time + duration, mesure, struct.pack(">BBB", 0x80 + channel.id, self._real_value, self.volume)),
      )
  return ()

def _note_2_midi_events_link(self, channel, mesure, start_time, end_time):
  real_time   = self.real_time()
  first_value = self.link_first_real_value()
  _min, _max  = self.link_min_max_real_values()
  delta = max(first_value - _min, _max - first_value)
  
  midi_events = [] # Here we'll use the "special ID" channel, and not the channel normal ID, to avoid to change other notes played in parallel on the normal channel
  
  duration = self.link_duration()
  if getattr(self.partition, "let_ring", 0):
    next = self.next()
    while next and next.time < self.time + duration:
      next = next.next()
    if next: duration = min(next.time - self.time, 384)
    else:    duration = 384
    
  if not self.linked_from:
    if delta:
      midi_set_pitch_bend_delta(midi_events, real_time - start_time, mesure, channel.special_id, delta)
      
    if (self.fx == "harmonic") and (not self.partition.view.use_harmonics_for_octavo):
      midi_events.append((real_time - start_time,            mesure, struct.pack(">BBB", 0x90 + channel.special_id, self._real_value - 12, self.volume), self))
      midi_events.append((real_time - start_time + duration, mesure, struct.pack(">BBB", 0x80 + channel.special_id, self._real_value - 12, self.volume)))
      midi_events.append((real_time - start_time,            mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, int(63.0 + 63.0 * 12.0 / delta)))) # Reset initial pitch bend
    else:
      midi_events.append((real_time - start_time,            mesure, struct.pack(">BBB", 0x90 + channel.special_id, self._real_value, self.volume), self))
      midi_events.append((real_time - start_time + duration, mesure, struct.pack(">BBB", 0x80 + channel.special_id, self._real_value, self.volume)))
      midi_events.append((real_time - start_time,            mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, 63))) # Reset initial pitch bend
      
  midi_events.append((real_time - start_time + duration - 1, mesure, struct.pack(">BBB", 0xB0 + channel.special_id, 0x7B, 127))) # Stop all sound on the special channel, because we are going to modify the pitch bend !
  
  if delta:
    if (self.linked_to.fx == "harmonic") and (not self.partition.view.use_harmonics_for_octavo):
      midi_events.append((self.linked_to.real_time() - start_time, mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, int(63.0 + 63.0 * (self.linked_to._real_value + 12.0 - self.link_first_real_value()) / delta)))) # Pitch bend
    else:
      midi_events.append((self.linked_to.real_time() - start_time, mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, int(63.0 + 63.0 * (self.linked_to._real_value - self.link_first_real_value()) / delta)))) # Pitch bend
      
  return tuple(midi_events)


def _note_2_midi_events_slide(self, channel, mesure, start_time, end_time):
  real_time   = self.real_time()
  first_value = self.link_first_real_value()
  _min, _max  = self.link_min_max_real_values()
  delta = max(first_value - _min, _max - first_value)
  
  midi_events = [] # Here we'll use the "special ID" channel, and not the channel normal ID, to avoid to change other notes played in parallel on the normal channel
  
  if not self.linked_from:
    if delta:
      midi_set_pitch_bend_delta(midi_events, real_time - start_time, mesure, channel.special_id, delta)
      
    midi_events.append((real_time - start_time                       , mesure, struct.pack(">BBB", 0x90 + channel.special_id, self._real_value, self.volume), self))
    midi_events.append((real_time - start_time + self.link_duration(), mesure, struct.pack(">BBB", 0x80 + channel.special_id, self._real_value, self.volume)))
    midi_events.append((real_time - start_time                       , mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, 63))) # Reset initial pitch bend
    
  delta_time  = self.linked_to.real_time() - real_time
  delta_value = self.linked_to._real_value - self.link_first_real_value()
  for i in range(1, 11):
    value = (i / 10.0) * self.linked_to._real_value + ((10 - i) / 10.0) * self._real_value
    f = i / 10.0
    midi_events.append((real_time + int(f * delta_time) - start_time,  mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, int(63.0 + 63.0 * (value - self.link_first_real_value()) / delta)))) # Pitch bend
    
  return tuple(midi_events)

def _note_2_midi_events_bend(self, channel, mesure, start_time, end_time):
  real_time   = self.real_time()
  midi_events = [] # Here we'll use the "special ID" channel, and not the channel normal ID, to avoid to change other notes played in parallel on the normal channel
  
  midi_events.append((real_time - start_time                       , mesure, struct.pack(">BBB", 0x90 + channel.special_id, self._real_value, self.volume), self))
  midi_events.append((real_time - start_time + self.link_duration(), mesure, struct.pack(">BBB", 0x80 + channel.special_id, self._real_value, self.volume)))
  midi_events.append((real_time - start_time                       , mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, 63))) # Reset initial pitch bend
  
  delta_time  = self.duration
  delta_value = 2.0 * self.bend_pitch
  for i in range(10):
    f = i / 10.0
    midi_events.append((real_time + int(f * delta_time) - start_time, mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, int(31.5 * (f * delta_value + 2))))) # Pitch bend
    
  return tuple(midi_events)

def _note_2_midi_events_tremolo(self, channel, mesure, start_time, end_time):
  real_time   = self.real_time()
  midi_events = [] # Here we'll use the "special ID" channel, and not the channel normal ID, to avoid to change other notes played in parallel on the normal channel
  
  midi_events.append((real_time - start_time,                        mesure, struct.pack(">BBB", 0x90 + channel.special_id, self._real_value, self.volume), self))
  midi_events.append((real_time - start_time + self.link_duration(), mesure, struct.pack(">BBB", 0x80 + channel.special_id, self._real_value, self.volume)))
  midi_events.append((real_time - start_time,                        mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, 63))) # Reset initial pitch bend
  
  val = 0.0
  for time in range(int(real_time - start_time), int(real_time - start_time + self.duration), 2):
    midi_events.append((time, mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, int(31.5 * (math.sin(val) * 0.3 + 2))))) # Pitch bend
    val += 0.3
    
  return tuple(midi_events)

def _note_2_midi_events_dead(self, channel, mesure, start_time, end_time):
  real_time   = self.real_time()
  midi_events = [] # Here we'll use the "special ID" channel, and not the channel normal ID, to avoid to change other notes played in parallel on the normal channel

  midi_events.append((real_time - start_time,                        mesure, struct.pack(">BBB", 0x90 + channel.special_id, self._real_value, self.volume), self))
  midi_events.append((real_time - start_time + self.link_duration(), mesure, struct.pack(">BBB", 0x80 + channel.special_id, self._real_value, self.volume)))
  midi_events.append((real_time - start_time,                        mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, 63))) # Reset initial pitch bend

  midi_events.append((real_time - start_time + 16,                   mesure, struct.pack(">BBB", 0xB0 + channel.special_id, 120, 0))) # Stop all sounds
  
  return tuple(midi_events)

def _note_2_midi_events_harmonic(self, channel, mesure, start_time, end_time):
  if self.partition.view.use_harmonics_for_octavo: return None
  if self.linked_from: return ()
  real_time = self.real_time()
  if getattr(self.partition, "let_ring", 0):
    next = self.next()
    if next: duration = min(next.time - self.time, 384)
    else:    duration = 384
  else: duration = self.duration
  
  midi_events = [] # Here we'll use the "special ID" channel, and not the channel normal ID, to avoid to change other notes played in parallel on the normal channel
  
  midi_set_pitch_bend_delta(midi_events, real_time - start_time, mesure, channel.special_id, 12)
  midi_events.append((real_time - start_time           , mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, 127)))
  
  midi_events.append((real_time - start_time + duration, mesure, struct.pack(">BBB", 0xE0 + channel.special_id, 63, 63))) # Reset pitch bend
  
  midi_events.append((real_time - start_time           , mesure, struct.pack(">BBB", 0x90 + channel.special_id, self._real_value - 12, self.volume), self))
  midi_events.append((real_time - start_time + duration, mesure, struct.pack(">BBB", 0x80 + channel.special_id, self._real_value - 12, self.volume)))
  
  return tuple(midi_events)

def midi_set_pitch_bend_delta(midi_events, time, mesure, channel_id, delta):
  if time >= 1:
    midi_events.append((time - 1, mesure, struct.pack(">BBB", 0xB0 + channel_id, 0x7B, 120))) # Stop all sound on the special channel, because we are going to modify the pitch bend !
  midi_events.append  ((time    , mesure, struct.pack(">BBB", 0xB0 + channel_id, 0x64, 0x00)))
  midi_events.append  ((time    , mesure, struct.pack(">BBB", 0xB0 + channel_id, 0x65, 0x00)))
  midi_events.append  ((time    , mesure, struct.pack(">BBB", 0xB0 + channel_id, 0x06, delta)))
  


def note_2_rich_midi_events(self):
  if not hasattr(self, "string_id"): return None
  if self.fx: return globals()["_note_2_rich_midi_events_%s" % self.fx](self)
  return b"\xff\x11\x01" + struct.pack(">b", self.string_id)

def _note_2_rich_midi_events_link(self):
  dif = cmp(self._real_value, self.linked_to._real_value)
  if   dif < 0: return b"\xff\x11\x02" + struct.pack(">b", self.string_id) + b"\x01"
  elif dif > 0: return b"\xff\x11\x02" + struct.pack(">b", self.string_id) + b"\x02"
  else:         return b"\xff\x11\x02" + struct.pack(">b", self.string_id) + b"\x01"  # Not supported by Rich Tablature Midi ???

def _note_2_rich_midi_events_slide(self):
  if self._real_value < self.linked_to._real_value: return b"\xff\x11\x02" + struct.pack(">b", self.string_id) + b"\x03"
  else:                                             return b"\xff\x11\x02" + struct.pack(">b", self.string_id) + b"\x04"

def _note_2_rich_midi_events_bend(self):
  return b"\xff\x11\x03" + struct.pack(">b", self.string_id) + b"\x0C" + struct.pack(">b", int(round(self.bend_pitch / 0.25)))

def _note_2_rich_midi_events_tremolo(self):
  return b"\xff\x11\x02" + struct.pack(">b", self.string_id) + b"\x0A"

def _note_2_rich_midi_events_dead(self):
  return b"\xff\x11\x02" + struct.pack(">b", self.string_id) + b"\x0E"

def _note_2_rich_midi_events_roll(self):
  notes = self.chord_notes()
  if (not notes) or (self._real_value == min([note._real_value for note in notes])):
    return b"\xff\x11\x02" + struct.pack(">b", self.string_id) + b"\x06"
  else:
    return b"\xff\x11\x01" + struct.pack(">b", self.string_id) # Same as normal notes
  
def _note_2_rich_midi_events_harmonic(self):
  return b"\xff\x11\x02" + struct.pack(">b", self.string_id) + b"\x07"



def chunk(name, data): return struct.pack(">4si", name, len(data)) + data

def varLength(i):
  array = [i & 0x7F]
  while i >= 128:
    i = i >> 7
    array.append((i & 0x7F) | 0x80)
  array.reverse()
  return struct.pack(">" + "B" * len(array), *array)

def readVarLength(file):
  result = 0
  while 1:
    b = struct.unpack(">B", file.read(1))[0]
    result = result | (b & 0x7F)
    if (b & 0x80) != 0x80: break
    result = result << 7
  return result

END_TRACK = b"\xff\x2f\x00"




#! /usr/bin/env python

# Songwrite 3
# Copyright (C) 2001-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, os.path, sys
import setuptools

HERE = os.path.dirname(sys.argv[0]) or "."

def do(command):
  print(command, file = sys.stderr)
  os.system(command)


package_data = {"songwrite3" : ["data/*",
                                "manpage/*", "manpage/man1/*",
                                "locale/*", "locale/en/*", "locale/en/LC_MESSAGES/*", "locale/fr/*", "locale/fr/LC_MESSAGES/*",
                                "doc/*", "doc/en/*", "doc/fr/*",
]}
  
if "--no-xdg" in sys.argv:
  sys.argv.remove("--no-xdg")
  no_xdg = 1
else:
  no_xdg = 0

if ("uninstall" in sys.argv) or ("uninstall-xdg" in sys.argv):
  print(file = sys.stderr)
  print("XDG un-installation (application, MIME type and icons)...", file = sys.stderr)
  do("xdg-mime uninstall %s/application-x-songwrite.xml" % HERE)
  do("xdg-icon-resource uninstall --novendor --context apps --noupdate --size 32 songwrite3" % HERE)
  do("xdg-icon-resource uninstall --novendor --context apps --noupdate --size 64 songwrite3" % HERE)
  do("xdg-desktop-menu uninstall --novendor %s/songwrite3.desktop" % HERE)
  sys.argv.remove("uninstall-xdg")

#data_files = [
#  (os.path.join("songwrite3", "data"),
#  [os.path.join("data", file) for file in os.listdir("data") if (file != "CVS") and (file != ".hg") and (file != ".svn") and (file != ".arch-ids")]),
#  (os.path.join("man", "man1"),
#  [os.path.join("manpage", "man1", "songwrite3")]),
#  ]
#if not no_lang:
#  data_files = data_files + [
#    (os.path.join(os.path.dirname(mo_file)), [mo_file])
#    for mo_file
#    in  glob.glob(os.path.join(".", "locale", "*", "LC_MESSAGES", "*"))
#    ]
#
#for doc_lang in ["fr", "en"]:
#  data_files.append(
#    (os.path.join("doc", "songwrite3", doc_lang),
#     glob.glob(os.path.join(".", "doc", doc_lang, "doc.pdf"))),
#    )

#distutils.core.setup(
setuptools.setup(
  name         = "Songwrite3",
  version      = "0.1",
  license      = "GPL",
  description  = "Tablature editor with Python 3, Qt 5, EditObj 3, and Timidity",
  author       = "Lamy Jean-Baptiste",
  author_email = "jibalamy@free.fr",
  url          = "http://www.lesfleursdunormal.fr/static/informatique/songwrite/index_en.html",
  
  package_dir  = {"songwrite3" : "."},
  packages     = ["songwrite3",
                  "songwrite3.plugins",
                  "songwrite3.plugins.abc",
                  "songwrite3.plugins.additional_instruments",
                  "songwrite3.plugins.fingering",
                  "songwrite3.plugins.lyre",
                  "songwrite3.plugins.midi",
                  "songwrite3.plugins.pdf",
                  "songwrite3.plugins.songwrite",
                  "songwrite3.plugins.texttab",
                  "songwrite3.plugins.chord",
                  "songwrite3.plugins.accordion",
  ],
  
  install_requires = ["Editobj3"], # XXX PyQt5 not supported here, but it IS required
  
  scripts      = ["songwrite3"],
  platforms    = ["Unix", "Linux"],
  
  classifiers  = [
    "Development Status :: 4 - Beta",
    "Environment :: X11 Applications",
    "Environment :: X11 Applications :: Qt",
    "Intended Audience :: End Users/Desktop",
    "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    "Natural Language :: French",
    "Natural Language :: English",
    "Operating System :: Unix",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.5",
    "Topic :: Artistic Software",
    "Topic :: Multimedia :: Sound/Audio",
    "Topic :: Multimedia :: Sound/Audio :: Editors",
    "Topic :: Multimedia :: Sound/Audio :: MIDI",
    ],
  
  package_data = package_data,
)


if ("install" in sys.argv) and (not no_xdg):
  print(file = sys.stderr)
  print("XDG installation (application, MIME type and icons)...", file = sys.stderr)
  do("xdg-mime install %s/application-x-songwrite.xml" % HERE)
  do("xdg-icon-resource install --novendor --context apps      --noupdate --size 32 %s/data/songwrite3_32x32.png songwrite3" % HERE)
  do("xdg-icon-resource install --novendor --context apps                 --size 64 %s/data/songwrite3_64x64.png songwrite3" % HERE)
  #do("xdg-icon-resource install            --context mimetypes --noupdate --size 32 %s/data/songwrite3_32x32.png application-x-songwrite" % HERE)
  #do("xdg-icon-resource install            --context mimetypes            --size 64 %s/data/songwrite3_64x64.png application-x-songwrite" % HERE)
  do("xdg-desktop-menu install  --novendor %s/songwrite3.desktop" % HERE)
